@@ -8,10 +8,12 @@ import (
 	"math/rand/v2"
 	"net/http"
 	"sync"
+	"sync/atomic"
 	"time"
 
 	"github.com/buildkite/agent/v3/api"
 	"github.com/buildkite/agent/v3/core"
+	"github.com/buildkite/agent/v3/internal/ptr"
 	"github.com/buildkite/agent/v3/logger"
 	"github.com/buildkite/agent/v3/metrics"
 	"github.com/buildkite/agent/v3/process"
@@ -57,7 +59,7 @@ type AgentWorker struct {
 	stats agentStats
 
 	// The API Client used when this agent is communicating with the API
-	apiClient APIClient
+	apiClient *api.Client
 
 	// The core Client is used to drive some APIClient methods
 	client *core.Client
@@ -86,25 +88,31 @@ type AgentWorker struct {
 	// The signal to use for cancellation
 	cancelSig process.Signal
 
-	// Stop controls
-	stopMutex sync.Mutex
-	stop      chan struct{}
-	stopping  bool
+	// Stop controls. Note that Stopping != Cancelling. See the [Stop] method
+	// for an explanation.
+	stopOnce sync.Once // prevents double-closing the channel
+	stop     chan struct{}
 
 	// The index of this agent worker
 	spawnIndex int
 
 	// When this worker runs a job, we'll store an instance of the
 	// JobRunner here
-	jobRunner jobRunner
+	jobRunner atomic.Pointer[JobRunner]
 
 	// Stdout of the parent agent process. Used for job log stdout writing arg, for simpler containerized log collection.
 	agentStdout io.Writer
 
 	// Are we doing something right now?
+	stateMtx     sync.Mutex
 	state        agentWorkerState
 	currentJobID string
-	stateMtx     sync.Mutex
+
+	// The time when this agent worker started
+	startTime time.Time
+
+	// disable the delay between pings, to speed up certain testing scenarios
+	noWaitBetweenPingsForTesting bool
 }
 
 type agentWorkerState string
@@ -165,11 +173,11 @@ func (e *errUnrecoverable) Unwrap() error {
 }
 
 // Creates the agent worker and initializes its API Client
-func NewAgentWorker(l logger.Logger, a *api.AgentRegisterResponse, m *metrics.Collector, apiClient APIClient, c AgentWorkerConfig) *AgentWorker {
-	apiClient = apiClient.FromAgentRegisterResponse(a)
+func NewAgentWorker(l logger.Logger, reg *api.AgentRegisterResponse, m *metrics.Collector, apiClient *api.Client, c AgentWorkerConfig) *AgentWorker {
+	apiClient = apiClient.FromAgentRegisterResponse(reg)
 	return &AgentWorker{
 		logger:           l,
-		agent:            a,
+		agent:            reg,
 		metricsCollector: m,
 		apiClient:        apiClient,
 		client: &core.Client{
@@ -209,19 +217,21 @@ func (a *AgentWorker) statusCallback(context.Context) (any, error) {
 }
 
 // Starts the agent worker
-func (a *AgentWorker) Start(ctx context.Context, idleMonitor *IdleMonitor) error {
+func (a *AgentWorker) Start(ctx context.Context, idleMon *idleMonitor) (startErr error) {
+	// Record the start time for max agent lifetime tracking
+	a.startTime = time.Now()
+
 	a.metrics = a.metricsCollector.Scope(metrics.Tags{
 		"agent_name": a.agent.Name,
 	})
 
-	ctx, done := status.AddItem(ctx, fmt.Sprintf("Worker %d", a.spawnIndex), workerStatusPart, a.statusCallback)
-	defer done()
+	ctx, _ = status.AddItem(ctx, fmt.Sprintf("Worker %d", a.spawnIndex), workerStatusPart, a.statusCallback)
 
 	// Start running our metrics collector
 	if err := a.metricsCollector.Start(); err != nil {
 		return err
 	}
-	defer a.metricsCollector.Stop()
+	defer a.metricsCollector.Stop() //nolint:errcheck // Best-effort cleanup
 
 	// Use a context to run heartbeats for as long as the ping loop or job runs
 	heartbeatCtx, cancel := context.WithCancel(ctx)
@@ -229,24 +239,43 @@ func (a *AgentWorker) Start(ctx context.Context, idleMonitor *IdleMonitor) error
 
 	go a.runHeartbeatLoop(heartbeatCtx)
 
-	// If the agent is booted in acquisition mode, then we don't need to
-	// bother about starting the ping loop.
+	// If the agent is booted in acquisition mode, acquire that particular job
+	// before running the ping loop.
+	// (Why run a ping loop at all? To find out if the agent is paused, which
+	// affects whether it terminates after the job.)
 	if a.agentConfiguration.AcquireJob != "" {
 		// When in acquisition mode, there can't be any agents, so
 		// there's really no point in letting the idle monitor know
 		// we're busy, but it's probably a good thing to do for good
 		// measure.
-		idleMonitor.MarkBusy(a.agent.UUID)
+		idleMon.markBusy(a)
 
-		return a.AcquireAndRunJob(ctx, a.agentConfiguration.AcquireJob)
+		if err := a.AcquireAndRunJob(ctx, a.agentConfiguration.AcquireJob); err != nil {
+			// If the job acquisition was rejected, we can exit with an error
+			// so that supervisor knows that the job was not acquired due to the job being rejected.
+			if errors.Is(err, core.ErrJobAcquisitionRejected) {
+				return fmt.Errorf("Failed to acquire job %q: %w", a.agentConfiguration.AcquireJob, err)
+			}
+
+			// If the job itself exited with nonzero code, then we want to exit
+			// with that code ourselves later on, but need to check if we were
+			// paused in the meantime first.
+			if exit := new(core.ProcessExit); errors.As(err, exit) && exit.Status != 0 {
+				defer func() {
+					startErr = errors.Join(err, startErr)
+				}()
+			}
+
+			a.logger.Error("Failed to acquire and run job: %v", err)
+		}
 	}
 
-	return a.runPingLoop(ctx, idleMonitor)
+	return a.runPingLoop(ctx, idleMon)
 }
 
 func (a *AgentWorker) runHeartbeatLoop(ctx context.Context) {
-	ctx, setStat, done := status.AddSimpleItem(ctx, "Heartbeat loop")
-	defer done()
+	ctx, setStat, _ := status.AddSimpleItem(ctx, "Heartbeat loop")
+	defer setStat("💔 Heartbeat loop stopped!")
 	setStat("🏃 Starting...")
 
 	heartbeatInterval := time.Second * time.Duration(a.agent.HeartbeatInterval)
@@ -282,9 +311,12 @@ func (a *AgentWorker) runHeartbeatLoop(ctx context.Context) {
 	}
 }
 
-func (a *AgentWorker) runPingLoop(ctx context.Context, idleMonitor *IdleMonitor) error {
-	ctx, setStat, done := status.AddSimpleItem(ctx, "Ping loop")
-	defer done()
+// runPingLoop runs the loop that pings Buildkite for work. It does all critical
+// agent things.
+// The lifetime of an agent is the lifetime of the ping loop.
+func (a *AgentWorker) runPingLoop(ctx context.Context, idleMon *idleMonitor) error {
+	ctx, setStat, _ := status.AddSimpleItem(ctx, "Ping loop")
+	defer setStat("🛑 Ping loop stopped!")
 	setStat("🏃 Starting...")
 
 	// Create the ticker
@@ -292,22 +324,47 @@ func (a *AgentWorker) runPingLoop(ctx context.Context, idleMonitor *IdleMonitor)
 	pingTicker := time.NewTicker(pingInterval)
 	defer pingTicker.Stop()
 
-	first := make(chan struct{}, 1)
-	first <- struct{}{}
+	// testTriggerCh will normally block forever, and so will not affect the for/select loop.
+	var testTriggerCh chan struct{}
+	if a.noWaitBetweenPingsForTesting {
+		// a closed channel will unblock the for/select instantly, for zero-delay ping loop testing.
+		testTriggerCh = make(chan struct{})
+		close(testTriggerCh)
+	}
 
-	lastActionTime := time.Now()
-	a.logger.Info("Waiting for work...")
+	// On the first iteration, skip waiting for the pingTicker.
+	// This doesn't skip the jitter, though.
+	skipTicker := make(chan struct{}, 1)
+	skipTicker <- struct{}{}
 
-	// Continue this loop until the closing of the stop channel signals termination
+	a.logger.Info("Waiting for instructions...")
+
+	ranJob := false
+	wasPaused := false
+
+	// Continue this loop until one of:
+	// * the context is cancelled
+	// * the stop channel is closed (a.Stop)
+	// * the agent is in acquire mode and the ping action isn't "pause"
+	// * the agent is in disconnect-after-job mode, the job is finished, and the
+	//   ping action isn't "pause",
+	// * the agent is in disconnect-after-idle-timeout mode, has been idle for
+	//   longer than the idle timeout, and the ping action isn't "pause".
+	// * the agent has exceeded its disconnect-after-uptime and the ping action isn't "pause".
 	for {
+		startWait := time.Now()
 		setStat("😴 Waiting until next ping interval tick")
 		select {
-		case <-first:
+		case <-testTriggerCh:
+			// instant receive from closed chan when noWaitBetweenPingsForTesting is true
+		case <-skipTicker:
 			// continue below
 		case <-pingTicker.C:
 			// continue below
 		case <-a.stop:
 			return nil
+		case <-ctx.Done():
+			return ctx.Err()
 		}
 
 		// Within the interval, wait a random amount of time to avoid
@@ -315,132 +372,182 @@ func (a *AgentWorker) runPingLoop(ctx context.Context, idleMonitor *IdleMonitor)
 		jitter := rand.N(pingInterval)
 		setStat(fmt.Sprintf("🫨 Jittering for %v", jitter))
 		select {
+		case <-testTriggerCh:
+			// instant receive from closed chan when noWaitBetweenPingsForTesting is true
 		case <-time.After(jitter):
 			// continue below
 		case <-a.stop:
 			return nil
+		case <-ctx.Done():
+			return ctx.Err()
 		}
+		pingWaitDurations.Observe(time.Since(startWait).Seconds())
 
-		a.stopMutex.Lock()
-		stopping := a.stopping
-		a.stopMutex.Unlock()
-		if stopping {
-			return nil
-		}
-
-		setStat("📡 Pinging Buildkite for work")
-		job, err := a.Ping(ctx)
+		setStat("📡 Pinging Buildkite for instructions")
+		pingsSent.Inc()
+		startPing := time.Now()
+		job, action, err := a.Ping(ctx)
 		if err != nil {
+			pingErrors.Inc()
 			if errors.Is(err, &errUnrecoverable{}) {
 				a.logger.Error("%v", err)
 			} else {
 				a.logger.Warn("%v", err)
 			}
-		} else if job != nil {
-			// Let other agents know this agent is now busy and
-			// not to idle terminate
-			idleMonitor.MarkBusy(a.agent.UUID)
+		}
+		pingDurations.Observe(time.Since(startPing).Seconds())
 
-			setStat("💼 Accepting job")
+		pingActions.WithLabelValues(action).Inc()
+		switch action {
+		case "disconnect":
+			a.StopUngracefully()
+			return nil
 
-			// Runs the job, only errors if something goes wrong
-			if runErr := a.AcceptAndRunJob(ctx, job); runErr != nil {
-				a.logger.Error("%v", runErr)
-			} else {
-				if a.agentConfiguration.DisconnectAfterJob {
-					a.logger.Info("Job finished. Disconnecting...")
-					return nil
-				}
-				lastActionTime = time.Now()
-
-				// Observation: jobs are rarely the last within a pipeline,
-				// thus if this worker just completed a job,
-				// there is likely another immediately available.
-				// Skip waiting for the ping interval until
-				// a ping without a job has occurred,
-				// but in exchange, ensure the next ping must wait a full
-				// pingInterval to avoid too much server load.
-
-				pingTicker.Reset(pingInterval)
-
-				continue
-			}
-			setStat("✅ Finished job")
+		case "pause":
+			// An agent is not dispatched any jobs while it is paused, but the
+			// paused agent is expected to remain alive and pinging for
+			// instructions.
+			// *This includes acquire-job and disconnect-after-idle-timeout.*
+			wasPaused = true
+			continue
 		}
 
-		// Handle disconnect after idle timeout (and deprecated disconnect-after-job-timeout)
-		if a.agentConfiguration.DisconnectAfterIdleTimeout > 0 {
-			idleDeadline := lastActionTime.Add(time.Second *
-				time.Duration(a.agentConfiguration.DisconnectAfterIdleTimeout))
+		// At this point, action was neither "disconnect" nor "pause".
+		if wasPaused {
+			a.logger.Info("Agent has resumed after being paused")
+			wasPaused = false
+		}
 
-			if time.Now().After(idleDeadline) {
-				// Let other agents know this agent is now idle and termination
-				// is possible
-				idleMonitor.MarkIdle(a.agent.UUID)
-
-				// But only terminate if everyone else is also idle
-				if idleMonitor.Idle() {
-					a.logger.Info("All agents have been idle for %d seconds. Disconnecting...",
-						a.agentConfiguration.DisconnectAfterIdleTimeout)
-					return nil
-				} else {
-					a.logger.Debug("Agent has been idle for %.f seconds, but other agents haven't",
-						time.Since(lastActionTime).Seconds())
-				}
+		// Exit after acquire-job.
+		// For acquire-job agents, registration sets ignore-in-dispatches=true,
+		// so job should be nil. If not nil, complain.
+		if a.agentConfiguration.AcquireJob != "" {
+			if job != nil {
+				a.logger.Error("Agent ping dispatched a job (id %q) but agent is in acquire-job mode!", job.ID)
 			}
+			return nil
+		}
+
+		// Exit after disconnect-after-job. Finishing the job sets
+		// ignore-in-dispatches=true, so job should be nil. If not, complain.
+		if ranJob && a.agentConfiguration.DisconnectAfterJob {
+			if job != nil {
+				a.logger.Error("Agent ping dispatched a job (id %q) but agent is in disconnect-after-job mode (and already ran a job)!", job.ID)
+			}
+			a.logger.Info("Job ran, and disconnect-after-job is enabled. Disconnecting...")
+			return nil
+		}
+
+		// Exit after disconnect-after-uptime is exceeded.
+		if maxUptime := a.agentConfiguration.DisconnectAfterUptime; maxUptime > 0 {
+			if time.Since(a.startTime) >= maxUptime {
+				if job != nil {
+					a.logger.Error("Agent ping dispatched a job (id %q) but agent has exceeded max uptime of %v!", job.ID, maxUptime)
+				}
+				a.logger.Info("Agent has exceeded max uptime of %v. Disconnecting...", maxUptime)
+				return nil
+			}
+		}
+
+		// Note that Ping only returns a job if err == nil.
+		if job == nil {
+			idleTimeout := a.agentConfiguration.DisconnectAfterIdleTimeout
+			if idleTimeout == 0 {
+				// No job and no idle timeout.
+				continue
+			}
+
+			// This ensures agents that never receive a job are still tracked
+			// by the idle monitor and can properly trigger disconnect-after-idle-timeout.
+			idleMon.markIdle(a)
+
+			// Exit if every agent has been idle for at least the timeout.
+			if idleMon.shouldExit(idleTimeout) {
+				a.logger.Info("All agents have been idle for at least %v. Disconnecting...", idleTimeout)
+				return nil
+			}
+
+			// Not idle enough to exit. Wait and ping again.
+			continue
+		}
+
+		setStat("💼 Accepting job")
+
+		// Runs the job, only errors if something goes wrong
+		if err := a.AcceptAndRunJob(ctx, job, idleMon); err != nil {
+			a.logger.Error("%v", err)
+			setStat(fmt.Sprintf("✅ Finished job with error: %v", err))
+			continue
+		}
+
+		ranJob = true
+
+		// Observation: jobs are rarely the last within a pipeline,
+		// thus if this worker just completed a job,
+		// there is likely another immediately available.
+		// Skip waiting for the ping interval until
+		// a ping without a job has occurred,
+		// but in exchange, ensure the next ping must wait at least a full
+		// pingInterval to avoid too much server load.
+		pingTicker.Reset(pingInterval)
+		select {
+		case skipTicker <- struct{}{}:
+			// Ticker will be skipped
+		default:
+			// We're already skipping the ticker, don't block.
 		}
 	}
 }
 
-// Stops the agent from accepting new work and cancels any current work it's
-// running
-func (a *AgentWorker) Stop(graceful bool) {
-	// Only allow one stop to run at a time (because we're playing with
-	// channels)
-	a.stopMutex.Lock()
-	defer a.stopMutex.Unlock()
+// StopGracefully stops the agent from accepting new work. It allows the current
+// job to finish without interruption. Does not block.
+func (a *AgentWorker) StopGracefully() {
+	select {
+	case <-a.stop:
+		a.logger.Warn("Agent is already gracefully stopping...")
+		return
 
-	if graceful {
-		if a.stopping {
-			a.logger.Warn("Agent is already gracefully stopping...")
-		} else {
-			// If we have a job, tell the user that we'll wait for
-			// it to finish before disconnecting
-			if a.jobRunner != nil {
-				a.logger.Info("Gracefully stopping agent. Waiting for current job to finish before disconnecting...")
-			} else {
-				a.logger.Info("Gracefully stopping agent. Since there is no job running, the agent will disconnect immediately")
-			}
+	default:
+		// continue below
+	}
+
+	// If we have a job, tell the user that we'll wait for it to finish
+	// before disconnecting
+	if a.jobRunner.Load() != nil {
+		a.logger.Info("Gracefully stopping agent. Waiting for current job to finish before disconnecting...")
+	} else {
+		a.logger.Info("Gracefully stopping agent. Since there is no job running, the agent will disconnect immediately")
+	}
+
+	a.stopOnce.Do(func() {
+		// Use the closure of the stop channel as a signal to the main run
+		// loop in Start() to stop looping and terminate
+		close(a.stop)
+	})
+}
+
+// StopUngracefully stops the agent from accepting new work and cancels any
+// existing job. It blocks until the job is cancelled, if there is one.
+func (a *AgentWorker) StopUngracefully() {
+	a.stopOnce.Do(func() {
+		// Use the closure of the stop channel as a signal to the main run
+		// loop in Start() to stop looping and terminate
+		close(a.stop)
+	})
+
+	// If there's a job running, kill it, then disconnect.
+	if jr := a.jobRunner.Load(); jr != nil {
+		a.logger.Info("Forcefully stopping agent. The current job will be canceled before disconnecting...")
+
+		// Kill the current job. Doesn't do anything if the job
+		// is already being killed, so it's safe to call
+		// multiple times.
+		if err := jr.Cancel(CancelReasonAgentStopping); err != nil {
+			a.logger.Error("Unexpected error canceling job (err: %s)", err)
 		}
 	} else {
-		// If there's a job running, kill it, then disconnect
-		if a.jobRunner != nil {
-			a.logger.Info("Forcefully stopping agent. The current job will be canceled before disconnecting...")
-
-			// Kill the current job. Doesn't do anything if the job
-			// is already being killed, so it's safe to call
-			// multiple times.
-			err := a.jobRunner.CancelAndStop()
-			if err != nil {
-				a.logger.Error("Unexpected error canceling job (err: %s)", err)
-			}
-		} else {
-			a.logger.Info("Forcefully stopping agent. Since there is no job running, the agent will disconnect immediately")
-		}
+		a.logger.Info("Forcefully stopping agent. Since there is no job running, the agent will disconnect immediately")
 	}
-
-	// We don't need to do the below operations again since we've already
-	// done them before
-	if a.stopping {
-		return
-	}
-
-	// Use the closure of the stop channel as a signal to the main run loop in Start()
-	// to stop looping and terminate
-	close(a.stop)
-
-	// Mark the agent as stopping
-	a.stopping = true
 }
 
 // Connects the agent to the Buildkite Agent API, retrying up to 10 times if it
@@ -451,7 +558,6 @@ func (a *AgentWorker) Connect(ctx context.Context) error {
 
 // Performs a heatbeat
 func (a *AgentWorker) Heartbeat(ctx context.Context) error {
-
 	// Retry the heartbeat a few times
 	r := roko.NewRetrier(
 		roko.WithMaxAttempts(10),
@@ -462,7 +568,7 @@ func (a *AgentWorker) Heartbeat(ctx context.Context) error {
 		b, resp, err := a.apiClient.Heartbeat(ctx)
 		if err != nil {
 			if resp != nil && !api.IsRetryableStatus(resp) {
-				a.Stop(false)
+				a.StopUngracefully()
 				r.Break()
 				return nil, &errUnrecoverable{action: "Heartbeat", response: resp, err: err}
 			}
@@ -491,7 +597,7 @@ func (a *AgentWorker) Heartbeat(ctx context.Context) error {
 
 // Performs a ping that checks Buildkite for a job or action to take
 // Returns a job, or nil if none is found
-func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
+func (a *AgentWorker) Ping(ctx context.Context) (job *api.Job, action string, err error) {
 	ping, resp, pingErr := a.apiClient.Ping(ctx)
 	// wait a minute, where's my if err != nil block? TL;DR look for pingErr ~20 lines down
 	// the api client returns an error if the response code isn't a 2xx, but there's still information in resp and ping
@@ -504,11 +610,7 @@ func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
 			a.logger.Info(ping.Message)
 		}
 
-		// Should the agent disconnect?
-		if ping.Action == "disconnect" {
-			a.Stop(false)
-			return nil, nil
-		}
+		action = ping.Action
 	}
 
 	if pingErr != nil {
@@ -516,8 +618,8 @@ func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
 		// The reason we do this after the disconnect check is because the backend can (and does) send disconnect actions in
 		// responses with non-retryable statuses
 		if resp != nil && !api.IsRetryableStatus(resp) {
-			a.Stop(false)
-			return nil, &errUnrecoverable{action: "Ping", response: resp, err: pingErr}
+			a.StopUngracefully()
+			return nil, action, &errUnrecoverable{action: "Ping", response: resp, err: pingErr}
 		}
 
 		// Get the last ping time to the nearest microsecond
@@ -527,9 +629,9 @@ func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
 		// If a ping fails, we don't really care, because it'll
 		// ping again after the interval.
 		if a.stats.lastPing.IsZero() {
-			return nil, fmt.Errorf("Failed to ping: %w (No successful ping yet)", pingErr)
+			return nil, action, fmt.Errorf("Failed to ping: %w (No successful ping yet)", pingErr)
 		} else {
-			return nil, fmt.Errorf("Failed to ping: %w (Last successful was %v ago)", pingErr, time.Since(a.stats.lastPing))
+			return nil, action, fmt.Errorf("Failed to ping: %w (Last successful was %v ago)", pingErr, time.Since(a.stats.lastPing))
 		}
 	}
 
@@ -556,11 +658,12 @@ func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
 	}
 
 	// If we don't have a job, there's nothing to do!
-	if ping.Job == nil {
-		return nil, nil
+	// If we're paused, job should be nil, but in case it isn't, ignore it.
+	if ping.Job == nil || action == "pause" {
+		return nil, action, nil
 	}
 
-	return ping.Job, nil
+	return ping.Job, action, nil
 }
 
 // AcquireAndRunJob attempts to acquire a job an run it. It will retry at after the
@@ -568,25 +671,37 @@ func (a *AgentWorker) Ping(ctx context.Context) (*api.Job, error) {
 // state. If the job is in an unassignable state, it will return an error immediately.
 // Otherwise, it will retry every 3s for 30 s. The whole operation will timeout after 5 min.
 func (a *AgentWorker) AcquireAndRunJob(ctx context.Context, jobId string) error {
-	ctx, cancel := context.WithCancel(ctx)
+	// Note: Context.Cancel is a blunt instrument. It will (for example)
+	// prevent the final API calls to upload remaining logs and mark the job
+	// finished.
+	// But we do want to abort the retry loop in AcquireJob early if possible.
+	// So, use context cancellation to abort AcquireJob on agent stop, but not
+	// RunJob.
+	// The agent's signal handler handles cancellation after a job has begun.
+	acquireCtx, cancel := context.WithCancel(ctx)
+	defer cancel()
 	go func() {
 		<-a.stop
 		cancel()
 	}()
 
-	job, err := a.client.AcquireJob(ctx, jobId)
+	job, err := a.client.AcquireJob(acquireCtx, jobId)
 	if err != nil {
 		return fmt.Errorf("failed to acquire job: %w", err)
 	}
 
-	// Now that we've acquired the job, let's run it
-	return a.RunJob(ctx, job)
+	// Now that we've acquired the job, let's run it.
+	return a.RunJob(ctx, job, nil)
 }
 
 // Accepts a job and runs it, only returns an error if something goes wrong
-func (a *AgentWorker) AcceptAndRunJob(ctx context.Context, job *api.Job) error {
+func (a *AgentWorker) AcceptAndRunJob(ctx context.Context, job *api.Job, idleMon *idleMonitor) error {
 	a.logger.Info("Assigned job %s. Accepting...", job.ID)
 
+	// An agent is busy during a job, and idle when the job is done.
+	idleMon.markBusy(a)
+	defer idleMon.markIdle(a)
+
 	// Accept the job. We'll retry on connection related issues, but if
 	// Buildkite returns a 422 or 500 for example, we'll just bail out,
 	// re-ping, and try the whole process again.
@@ -613,14 +728,24 @@ func (a *AgentWorker) AcceptAndRunJob(ctx context.Context, job *api.Job) error {
 		return fmt.Errorf("Failed to accept job: %w", err)
 	}
 
+	// If we're disconnecting-after-job, signal back to Buildkite that we're not
+	// interested in jobs after this one.
+	var ignoreAgentInDispatches *bool
+	if a.agentConfiguration.DisconnectAfterJob {
+		ignoreAgentInDispatches = ptr.To(true)
+	}
+
 	// Now that we've accepted the job, let's run it
-	return a.RunJob(ctx, accepted)
+	return a.RunJob(ctx, accepted, ignoreAgentInDispatches)
 }
 
-func (a *AgentWorker) RunJob(ctx context.Context, acceptResponse *api.Job) error {
+func (a *AgentWorker) RunJob(ctx context.Context, acceptResponse *api.Job, ignoreAgentInDispatches *bool) error {
 	a.setBusy(acceptResponse.ID)
 	defer a.setIdle()
 
+	jobsStarted.Inc()
+	defer jobsEnded.Inc()
+
 	jobMetricsScope := a.metrics.With(metrics.Tags{
 		"pipeline": acceptResponse.Env["BUILDKITE_PIPELINE_SLUG"],
 		"org":      acceptResponse.Env["BUILDKITE_ORGANIZATION_SLUG"],
@@ -645,14 +770,14 @@ func (a *AgentWorker) RunJob(ctx context.Context, acceptResponse *api.Job) error
 	if err != nil {
 		return fmt.Errorf("Failed to initialize job: %w", err)
 	}
-	a.jobRunner = jr
-	defer func() {
-		// No more job, no more runner.
-		a.jobRunner = nil
-	}()
+	if !a.jobRunner.CompareAndSwap(nil, jr) {
+		return fmt.Errorf("Agent worker already has a job running")
+	}
+	// No more job, no more runner.
+	defer a.jobRunner.Store(nil)
 
 	// Start running the job
-	if err := jr.Run(ctx); err != nil {
+	if err := jr.Run(ctx, ignoreAgentInDispatches); err != nil {
 		return fmt.Errorf("Failed to run job: %w", err)
 	}
 
@@ -683,44 +808,3 @@ func (a *AgentWorker) healthHandler() http.HandlerFunc {
 		}
 	}
 }
-
-// This monitor has a 3rd implicit state we will call "initializing" that all agents start in
-// Agents can transition to busy and/or idle but always start in the "initializing" state
-/*
-//                -> Busy
-//              /     ^
-// Initializing       |
-//              \     v
-//                -> Idle
-*/
-// This (intentionally?) ensures the DisconnectAfterIdleTimeout doesn't fire before agents have had a chance to run a job
-type IdleMonitor struct {
-	sync.Mutex
-	totalAgents int
-	idle        map[string]struct{}
-}
-
-func NewIdleMonitor(totalAgents int) *IdleMonitor {
-	return &IdleMonitor{
-		totalAgents: totalAgents,
-		idle:        map[string]struct{}{},
-	}
-}
-
-func (i *IdleMonitor) Idle() bool {
-	i.Lock()
-	defer i.Unlock()
-	return len(i.idle) == i.totalAgents
-}
-
-func (i *IdleMonitor) MarkIdle(agentUUID string) {
-	i.Lock()
-	defer i.Unlock()
-	i.idle[agentUUID] = struct{}{}
-}
-
-func (i *IdleMonitor) MarkBusy(agentUUID string) {
-	i.Lock()
-	defer i.Unlock()
-	delete(i.idle, agentUUID)
-}
