@@ -1,7 +1,6 @@
 package integration
 
 import (
-	"context"
 	"strings"
 	"testing"
 
@@ -46,6 +45,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "1", // step env overrides pipeline env
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithNoPluginConfig = api.Job{
@@ -67,6 +67,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "1", // step env overrides pipeline env
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithNoPlugins = api.Job{
@@ -82,6 +83,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithNullPlugins = api.Job{
@@ -97,6 +99,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMismatchedStepAndJob = api.Job{
@@ -110,6 +113,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMismatchedPlugins = api.Job{
@@ -130,6 +134,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMissingPlugins = api.Job{
@@ -150,6 +155,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMismatchedEnv = api.Job{
@@ -163,6 +169,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "crimes",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithStepEnvButNoCorrespondingJobEnv = api.Job{
@@ -177,6 +184,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithPipelineEnvButNoCorrespondingJobEnv = api.Job{
@@ -189,6 +197,7 @@ var (
 			"BUILDKITE_COMMAND": "echo hello world",
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMatrix = api.Job{
@@ -206,6 +215,7 @@ var (
 			"greeting": "hello",
 			"object":   "world",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithInvalidMatrixPermutation = api.Job{
@@ -223,6 +233,7 @@ var (
 			"greeting": "goodbye",
 			"object":   "mister anderson",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithMatrixMismatch = api.Job{
@@ -240,6 +251,7 @@ var (
 			"greeting": "hello",
 			"object":   "world",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithPipelineInvariantsInEnv = api.Job{
@@ -253,6 +265,7 @@ var (
 			"BUILDKITE_REPO":    defaultRepositoryURL,
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
 	}
 
 	jobWithInvalidPipelineInvariantsInEnv = api.Job{
@@ -266,12 +279,96 @@ var (
 			"BUILDKITE_REPO":    "https://github.com/haxors/agent.git",
 			"DEPLOY":            "0",
 		},
+		Token: "bkaj_job-token",
+	}
+
+	jobWithSecrets = api.Job{
+		ChunksMaxSizeBytes: 1024,
+		ID:                 defaultJobID,
+		Step: pipeline.CommandStep{
+			Command: "echo hello secrets",
+			Secrets: pipeline.Secrets{
+				pipeline.Secret{Key: "DATABASE_URL", EnvironmentVariable: "DATABASE_URL"},
+				pipeline.Secret{Key: "API_TOKEN", EnvironmentVariable: "API_TOKEN"},
+			},
+		},
+		Env: map[string]string{
+			"BUILDKITE_COMMAND": "echo hello secrets",
+			"BUILDKITE_REPO":    defaultRepositoryURL,
+			"DEPLOY":            "0",
+		},
+		Token: "bkaj_job-token",
+	}
+
+	jobWithSecretsCustomEnv = api.Job{
+		ChunksMaxSizeBytes: 1024,
+		ID:                 defaultJobID,
+		Step: pipeline.CommandStep{
+			Command: "echo hello custom secrets",
+			Secrets: pipeline.Secrets{
+				pipeline.Secret{Key: "DATABASE_URL", EnvironmentVariable: "DB_CONNECTION_STRING"},
+				pipeline.Secret{Key: "API_TOKEN", EnvironmentVariable: "CUSTOM_API_KEY"},
+			},
+		},
+		Env: map[string]string{
+			"BUILDKITE_COMMAND": "echo hello custom secrets",
+			"BUILDKITE_REPO":    defaultRepositoryURL,
+			"DEPLOY":            "0",
+		},
+		Token: "bkaj_job-token",
+	}
+
+	jobWithEmptySecrets = api.Job{
+		ChunksMaxSizeBytes: 1024,
+		ID:                 defaultJobID,
+		Step: pipeline.CommandStep{
+			Command: "echo hello no secrets",
+			Secrets: pipeline.Secrets{},
+		},
+		Env: map[string]string{
+			"BUILDKITE_COMMAND": "echo hello no secrets",
+			"BUILDKITE_REPO":    defaultRepositoryURL,
+			"DEPLOY":            "0",
+		},
+		Token: "bkaj_job-token",
+	}
+
+	jobWithNilSecrets = api.Job{
+		ChunksMaxSizeBytes: 1024,
+		ID:                 defaultJobID,
+		Step: pipeline.CommandStep{
+			Command: "echo hello no secrets",
+			Secrets: nil,
+		},
+		Env: map[string]string{
+			"BUILDKITE_COMMAND": "echo hello no secrets",
+			"BUILDKITE_REPO":    defaultRepositoryURL,
+			"DEPLOY":            "0",
+		},
+		Token: "bkaj_job-token",
+	}
+
+	jobWithMismatchedSecrets = api.Job{
+		ChunksMaxSizeBytes: 1024,
+		ID:                 defaultJobID,
+		Step: pipeline.CommandStep{
+			Command: "echo hello world",
+			Secrets: pipeline.Secrets{
+				pipeline.Secret{Key: "DATABASE_URL", EnvironmentVariable: "DATABASE_URL"},
+			},
+		},
+		Env: map[string]string{
+			"BUILDKITE_COMMAND": "echo hello world",
+			"BUILDKITE_REPO":    defaultRepositoryURL,
+			"DEPLOY":            "0",
+		},
+		Token: "bkaj_job-token",
 	}
 )
 
 func TestJobVerification(t *testing.T) {
 	t.Parallel()
-	ctx := context.Background()
+	ctx := t.Context()
 
 	cases := []struct {
 		name                     string
@@ -423,7 +520,7 @@ func TestJobVerification(t *testing.T) {
 		},
 		{
 			name:                     "when the step has a signature, but the JobRunner doesn't have a verification key, it fails signature verification",
-			agentConf:                agent.AgentConfiguration{},
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourBlock},
 			job:                      job,
 			repositoryURL:            defaultRepositoryURL,
 			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
@@ -433,7 +530,22 @@ func TestJobVerification(t *testing.T) {
 			expectedSignalReason:     agent.SignalReasonSignatureRejected,
 			expectLogsContain: []string{
 				"+++ ⛔",
-				"but no verification key was provided",
+				"cannot verify signature. JWK for pipeline verification is not configured",
+			},
+		},
+		{
+			name:                     "when the step has a signature, but the JobRunner doesn't have a verification key, and JobVerificationFailureBehaviour is warn, it warns and runs the job",
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourWarn},
+			job:                      job,
+			repositoryURL:            defaultRepositoryURL,
+			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
+			verificationJWKS:         nil,
+			mockBootstrapExpectation: func(bt *bintest.Mock) { bt.Expect().Once().AndExitWith(0) },
+			expectedExitStatus:       "0",
+			expectedSignalReason:     "",
+			expectLogsContain: []string{
+				"+++ ⛔",
+				"cannot verify signature. JWK for pipeline verification is not configured",
 			},
 		},
 		{
@@ -548,6 +660,50 @@ func TestJobVerification(t *testing.T) {
 				"signature verification failed",
 			},
 		},
+		{
+			name:                     "when job signature is valid and has secrets, it runs the job",
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourBlock},
+			job:                      jobWithSecrets,
+			repositoryURL:            defaultRepositoryURL,
+			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
+			verificationJWKS:         jwksFromKeys(t, symmetricJWKFor(t, signingKeyLlamas)),
+			mockBootstrapExpectation: func(bt *bintest.Mock) { bt.Expect().Once().AndExitWith(0) },
+			expectedExitStatus:       "0",
+			expectedSignalReason:     "",
+		},
+		{
+			name:                     "when job signature is valid and has no secrets, it runs the job",
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourBlock},
+			job:                      jobWithNilSecrets,
+			repositoryURL:            defaultRepositoryURL,
+			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
+			verificationJWKS:         jwksFromKeys(t, symmetricJWKFor(t, signingKeyLlamas)),
+			mockBootstrapExpectation: func(bt *bintest.Mock) { bt.Expect().Once().AndExitWith(0) },
+			expectedExitStatus:       "0",
+			expectedSignalReason:     "",
+		},
+		{
+			name:                     "when job signature is valid and has secrets with custom environment variables, it runs the job",
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourBlock},
+			job:                      jobWithSecretsCustomEnv,
+			repositoryURL:            defaultRepositoryURL,
+			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
+			verificationJWKS:         jwksFromKeys(t, symmetricJWKFor(t, signingKeyLlamas)),
+			mockBootstrapExpectation: func(bt *bintest.Mock) { bt.Expect().Once().AndExitWith(0) },
+			expectedExitStatus:       "0",
+			expectedSignalReason:     "",
+		},
+		{
+			name:                     "when job signature is valid and has empty secrets, it runs the job",
+			agentConf:                agent.AgentConfiguration{VerificationFailureBehaviour: agent.VerificationBehaviourBlock},
+			job:                      jobWithEmptySecrets,
+			repositoryURL:            defaultRepositoryURL,
+			signingKey:               symmetricJWKFor(t, signingKeyLlamas),
+			verificationJWKS:         jwksFromKeys(t, symmetricJWKFor(t, signingKeyLlamas)),
+			mockBootstrapExpectation: func(bt *bintest.Mock) { bt.Expect().Once().AndExitWith(0) },
+			expectedExitStatus:       "0",
+			expectedSignalReason:     "",
+		},
 	}
 
 	for _, tc := range cases {
@@ -561,14 +717,22 @@ func TestJobVerification(t *testing.T) {
 
 			mb := mockBootstrap(t)
 			tc.mockBootstrapExpectation(mb)
-			defer mb.CheckAndClose(t)
+			defer mb.CheckAndClose(t) //nolint:errcheck // bintest logs to t
 
-			stepWithInvariants := signature.CommandStepWithInvariants{
-				CommandStep:   tc.job.Step,
-				RepositoryURL: tc.repositoryURL,
+			t.Logf("%s: signing step with key: %v", t.Name(), tc.signingKey)
+			if tc.signingKey != nil {
+				err := signature.SignSteps(
+					ctx,
+					pipeline.Steps{&tc.job.Step},
+					tc.signingKey,
+					tc.repositoryURL,
+					signature.WithEnv(pipelineUploadEnv),
+				)
+				if err != nil {
+					t.Fatalf("signing step: %v", err)
+				}
 			}
 
-			tc.job.Step = signStep(t, ctx, tc.signingKey, pipelineUploadEnv, stepWithInvariants)
 			err := runJob(t, ctx, testRunJobConfig{
 				job:              &tc.job,
 				server:           server,
@@ -656,25 +820,3 @@ func jwksFromKeys(t *testing.T, jwkes ...jwk.Key) jwk.Set {
 
 	return set
 }
-
-func signStep(
-	t *testing.T,
-	ctx context.Context,
-	key jwk.Key,
-	env map[string]string,
-	stepWithInvariants signature.CommandStepWithInvariants,
-) pipeline.CommandStep {
-	t.Helper()
-
-	t.Logf("%s: signing step with key: %v", t.Name(), key)
-	if key == nil {
-		return stepWithInvariants.CommandStep
-	}
-
-	signature, err := signature.Sign(ctx, key, &stepWithInvariants, signature.WithEnv(env))
-	if err != nil {
-		t.Fatalf("signing step: %v", err)
-	}
-	stepWithInvariants.CommandStep.Signature = signature
-	return stepWithInvariants.CommandStep
-}
