@@ -16,7 +16,7 @@ import (
 const defaultLogMaxSize = 1024 * 1024 * 1024 // 1 GiB
 
 // Returned from Process after Stop has been called.
-var errStreamerStopped = errors.New("streamer stopped")
+var errStreamerStopped = errors.New("streamer was already stopped")
 
 // LogStreamerConfig contains configuration options for the log streamer.
 type LogStreamerConfig struct {
@@ -127,7 +127,7 @@ func (ls *LogStreamer) Process(ctx context.Context, output []byte) error {
 				humanize.IBytes(ls.bytes), humanize.IBytes(ls.conf.MaxSizeBytes))
 			ls.warnedAboutSize = true
 			// In a future version, this will error out, e.g.:
-			//return fmt.Errorf("%w (%d > %d)", errLogExceededMaxSize, ls.bytes, ls.conf.MaxSizeBytes)
+			// return fmt.Errorf("%w (%d > %d)", errLogExceededMaxSize, ls.bytes, ls.conf.MaxSizeBytes)
 		}
 
 		// The next chunk will be up to MaxChunkSizeBytes in size.
