@@ -8,7 +8,6 @@ import (
 	"io"
 	"maps"
 	"os"
-	"os/exec"
 	"os/signal"
 	"path/filepath"
 	"regexp"
@@ -30,7 +29,6 @@ import (
 	"github.com/buildkite/agent/v3/internal/awslib"
 	awssigner "github.com/buildkite/agent/v3/internal/cryptosigner/aws"
 	"github.com/buildkite/agent/v3/internal/experiments"
-	"github.com/buildkite/agent/v3/internal/job"
 	"github.com/buildkite/agent/v3/internal/job/hook"
 	"github.com/buildkite/agent/v3/internal/osutil"
 	"github.com/buildkite/agent/v3/internal/shell"
@@ -63,8 +61,6 @@ Example:
     $ buildkite-agent start --token xxx`
 
 var (
-	minGracePeriod = 10
-
 	verificationFailureBehaviors = []string{agent.VerificationBehaviourBlock, agent.VerificationBehaviourWarn}
 
 	buildkiteSetEnvironmentVariables = []*regexp.Regexp{
@@ -81,6 +77,8 @@ var (
 // - Into clicommand/bootstrap.go to read it from the env into the bootstrap config
 
 type AgentStartConfig struct {
+	GlobalConfig
+
 	Config string `cli:"config"`
 
 	Name              string   `cli:"name"`
@@ -103,8 +101,10 @@ type AgentStartConfig struct {
 	AcquireJob                 string `cli:"acquire-job"`
 	DisconnectAfterJob         bool   `cli:"disconnect-after-job"`
 	DisconnectAfterIdleTimeout int    `cli:"disconnect-after-idle-timeout"`
+	DisconnectAfterUptime      int    `cli:"disconnect-after-uptime"`
 	CancelGracePeriod          int    `cli:"cancel-grace-period"`
 	SignalGracePeriodSeconds   int    `cli:"signal-grace-period-seconds"`
+	ReflectExitStatus          bool   `cli:"reflect-exit-status"`
 
 	EnableJobLogTmpfile bool   `cli:"enable-job-log-tmpfile"`
 	JobLogPath          string `cli:"job-log-path" normalize:"filepath"`
@@ -151,14 +151,15 @@ type AgentStartConfig struct {
 	GitMirrorsSkipUpdate  bool   `cli:"git-mirrors-skip-update"`
 	NoGitSubmodules       bool   `cli:"no-git-submodules"`
 
-	NoSSHKeyscan        bool     `cli:"no-ssh-keyscan"`
-	NoCommandEval       bool     `cli:"no-command-eval"`
-	NoLocalHooks        bool     `cli:"no-local-hooks"`
-	NoPlugins           bool     `cli:"no-plugins"`
-	NoPluginValidation  bool     `cli:"no-plugin-validation"`
-	NoFeatureReporting  bool     `cli:"no-feature-reporting"`
-	AllowedRepositories []string `cli:"allowed-repositories" normalize:"list"`
-	AllowedPlugins      []string `cli:"allowed-plugins" normalize:"list"`
+	NoSSHKeyscan            bool     `cli:"no-ssh-keyscan"`
+	NoCommandEval           bool     `cli:"no-command-eval"`
+	NoLocalHooks            bool     `cli:"no-local-hooks"`
+	NoPlugins               bool     `cli:"no-plugins"`
+	NoPluginValidation      bool     `cli:"no-plugin-validation"`
+	PluginsAlwaysCloneFresh bool     `cli:"plugins-always-clone-fresh"`
+	NoFeatureReporting      bool     `cli:"no-feature-reporting"`
+	AllowedRepositories     []string `cli:"allowed-repositories" normalize:"list"`
+	AllowedPlugins          []string `cli:"allowed-plugins" normalize:"list"`
 
 	EnableEnvironmentVariableAllowList bool     `cli:"enable-environment-variable-allowlist"`
 	AllowedEnvironmentVariables        []string `cli:"allowed-environment-variables" normalize:"list"`
@@ -171,19 +172,15 @@ type AgentStartConfig struct {
 	MetricsDatadogDistributions bool   `cli:"metrics-datadog-distributions"`
 
 	// Tracing config
-	TracingBackend     string `cli:"tracing-backend"`
-	TracingServiceName string `cli:"tracing-service-name"`
+	TracingBackend              string `cli:"tracing-backend"`
+	TracingServiceName          string `cli:"tracing-service-name"`
+	TracingPropagateTraceparent bool   `cli:"tracing-propagate-traceparent"`
 
-	// Global flags
-	Debug                     bool     `cli:"debug"`
-	LogLevel                  string   `cli:"log-level"`
-	NoColor                   bool     `cli:"no-color"`
-	Experiments               []string `cli:"experiment" normalize:"list"`
-	Profile                   string   `cli:"profile"`
-	StrictSingleHooks         bool     `cli:"strict-single-hooks"`
-	KubernetesExec            bool     `cli:"kubernetes-exec"`
-	TraceContextEncoding      string   `cli:"trace-context-encoding"`
-	NoMultipartArtifactUpload bool     `cli:"no-multipart-artifact-upload"`
+	// Other shared flags
+	StrictSingleHooks         bool   `cli:"strict-single-hooks"`
+	KubernetesExec            bool   `cli:"kubernetes-exec"`
+	TraceContextEncoding      string `cli:"trace-context-encoding"`
+	NoMultipartArtifactUpload bool   `cli:"no-multipart-artifact-upload"`
 
 	// API config
 	DebugHTTP bool   `cli:"debug-http"`
@@ -193,14 +190,15 @@ type AgentStartConfig struct {
 	NoHTTP2   bool   `cli:"no-http2"`
 
 	// Deprecated
-	NoSSHFingerprintVerification bool     `cli:"no-automatic-ssh-fingerprint-verification" deprecated-and-renamed-to:"NoSSHKeyscan"`
-	MetaData                     []string `cli:"meta-data" deprecated-and-renamed-to:"Tags"`
-	MetaDataEC2                  bool     `cli:"meta-data-ec2" deprecated-and-renamed-to:"TagsFromEC2"`
-	MetaDataEC2Tags              bool     `cli:"meta-data-ec2-tags" deprecated-and-renamed-to:"TagsFromEC2Tags"`
-	MetaDataGCP                  bool     `cli:"meta-data-gcp" deprecated-and-renamed-to:"TagsFromGCP"`
-	TagsFromEC2                  bool     `cli:"tags-from-ec2" deprecated-and-renamed-to:"TagsFromEC2MetaData"`
-	TagsFromGCP                  bool     `cli:"tags-from-gcp" deprecated-and-renamed-to:"TagsFromGCPMetaData"`
-	DisconnectAfterJobTimeout    int      `cli:"disconnect-after-job-timeout" deprecated:"Use disconnect-after-idle-timeout instead"`
+	KubernetesLogCollectionGracePeriod time.Duration `cli:"kubernetes-log-collection-grace-period"`
+	NoSSHFingerprintVerification       bool          `cli:"no-automatic-ssh-fingerprint-verification" deprecated-and-renamed-to:"NoSSHKeyscan"`
+	MetaData                           []string      `cli:"meta-data" deprecated-and-renamed-to:"Tags"`
+	MetaDataEC2                        bool          `cli:"meta-data-ec2" deprecated-and-renamed-to:"TagsFromEC2"`
+	MetaDataEC2Tags                    bool          `cli:"meta-data-ec2-tags" deprecated-and-renamed-to:"TagsFromEC2Tags"`
+	MetaDataGCP                        bool          `cli:"meta-data-gcp" deprecated-and-renamed-to:"TagsFromGCP"`
+	TagsFromEC2                        bool          `cli:"tags-from-ec2" deprecated-and-renamed-to:"TagsFromEC2MetaData"`
+	TagsFromGCP                        bool          `cli:"tags-from-gcp" deprecated-and-renamed-to:"TagsFromGCPMetaData"`
+	DisconnectAfterJobTimeout          int           `cli:"disconnect-after-job-timeout" deprecated:"Use disconnect-after-idle-timeout instead"`
 }
 
 func (asc AgentStartConfig) Features(ctx context.Context) []string {
@@ -226,6 +224,10 @@ func (asc AgentStartConfig) Features(ctx context.Context) []string {
 		features = append(features, "opentelemetry-tracing")
 	}
 
+	if asc.TracingPropagateTraceparent {
+		features = append(features, "propagate-traceparent")
+	}
+
 	if asc.DisconnectAfterJob {
 		features = append(features, "disconnect-after-job")
 	}
@@ -234,6 +236,10 @@ func (asc AgentStartConfig) Features(ctx context.Context) []string {
 		features = append(features, "disconnect-after-idle")
 	}
 
+	if asc.DisconnectAfterUptime != 0 {
+		features = append(features, "disconnect-after-uptime")
+	}
+
 	if asc.NoPlugins {
 		features = append(features, "no-plugins")
 	}
@@ -326,7 +332,7 @@ var AgentStartCommand = cli.Command{
 	Name:        "start",
 	Usage:       "Starts a Buildkite agent",
 	Description: startDescription,
-	Flags: []cli.Flag{
+	Flags: append(globalFlags(),
 		cli.StringFlag{
 			Name:   "config",
 			Value:  "",
@@ -351,6 +357,11 @@ var AgentStartCommand = cli.Command{
 			Usage:  "Start this agent and only run the specified job, disconnecting after it's finished",
 			EnvVar: "BUILDKITE_AGENT_ACQUIRE_JOB",
 		},
+		cli.BoolFlag{
+			Name:   "reflect-exit-status",
+			Usage:  "When used with --acquire-job, causes the agent to exit with the same exit status as the job",
+			EnvVar: "BUILDKITE_AGENT_REFLECT_EXIT_STATUS",
+		},
 		cli.BoolFlag{
 			Name:   "disconnect-after-job",
 			Usage:  "Disconnect the agent after running exactly one job. When used in conjunction with the ′--spawn′ flag, each worker booted will run exactly one job",
@@ -362,6 +373,12 @@ var AgentStartCommand = cli.Command{
 			Usage:  "The maximum idle time in seconds to wait for a job before disconnecting. The default of 0 means no timeout",
 			EnvVar: "BUILDKITE_AGENT_DISCONNECT_AFTER_IDLE_TIMEOUT",
 		},
+		cli.IntFlag{
+			Name:   "disconnect-after-uptime",
+			Value:  0,
+			Usage:  "The maximum uptime in seconds before the agent stops accepting new jobs and shuts down after any running jobs complete. The default of 0 means no timeout",
+			EnvVar: "BUILDKITE_AGENT_DISCONNECT_AFTER_UPTIME",
+		},
 		cancelGracePeriodFlag,
 		cli.BoolFlag{
 			Name:   "enable-job-log-tmpfile",
@@ -538,12 +555,7 @@ var AgentStartCommand = cli.Command{
 			Usage:  "Additional directories to look for agent hooks",
 			EnvVar: "BUILDKITE_ADDITIONAL_HOOKS_PATHS",
 		},
-		cli.StringFlag{
-			Name:   "sockets-path",
-			Value:  defaultSocketsPath(),
-			Usage:  "Directory where the agent will place sockets",
-			EnvVar: "BUILDKITE_SOCKETS_PATH",
-		},
+		SocketsPathFlag,
 		cli.StringFlag{
 			Name:   "plugins-path",
 			Value:  "",
@@ -590,6 +602,11 @@ var AgentStartCommand = cli.Command{
 			Usage:  "Don't validate plugin configuration and requirements",
 			EnvVar: "BUILDKITE_NO_PLUGIN_VALIDATION",
 		},
+		cli.BoolFlag{
+			Name:   "plugins-always-clone-fresh",
+			Usage:  "Always make a new clone of plugin source, even if already present",
+			EnvVar: "BUILDKITE_PLUGINS_ALWAYS_CLONE_FRESH",
+		},
 		cli.BoolFlag{
 			Name:   "no-local-hooks",
 			Usage:  "Don't allow local hooks to be run from checked out repositories",
@@ -675,6 +692,11 @@ var AgentStartCommand = cli.Command{
 			EnvVar: "BUILDKITE_TRACING_BACKEND",
 			Value:  "",
 		},
+		cli.BoolFlag{
+			Name:   "tracing-propagate-traceparent",
+			Usage:  `Enable accepting traceparent context from Buildkite control plane (only supported for OpenTelemetry backend)`,
+			EnvVar: "BUILDKITE_TRACING_PROPAGATE_TRACEPARENT",
+		},
 		cli.StringFlag{
 			Name:   "tracing-service-name",
 			Usage:  "Service name to use when reporting traces.",
@@ -719,18 +741,13 @@ var AgentStartCommand = cli.Command{
 		},
 
 		// API Flags
-		AgentRegisterTokenFlag,
+		AgentRegisterTokenFlag, // != AgentAccessToken
 		EndpointFlag,
 		NoHTTP2Flag,
 		DebugHTTPFlag,
 		TraceHTTPFlag,
 
-		// Global flags
-		NoColorFlag,
-		DebugFlag,
-		LogLevelFlag,
-		ExperimentsFlag,
-		ProfileFlag,
+		// Other shared flags
 		RedactedVars,
 		StrictSingleHooksFlag,
 		KubernetesExecFlag,
@@ -738,6 +755,7 @@ var AgentStartCommand = cli.Command{
 		NoMultipartArtifactUploadFlag,
 
 		// Deprecated flags which will be removed in v4
+		KubernetesLogCollectionGracePeriodFlag,
 		cli.StringSliceFlag{
 			Name:   "meta-data",
 			Value:  &cli.StringSlice{},
@@ -752,7 +770,7 @@ var AgentStartCommand = cli.Command{
 		cli.BoolFlag{
 			Name:   "meta-data-ec2-tags",
 			Hidden: true,
-			EnvVar: "BUILDKITE_AGENT_TAGS_FROM_EC2_TAGS",
+			EnvVar: "BUILDKITE_AGENT_META_DATA_EC2_TAGS",
 		},
 		cli.BoolFlag{
 			Name:   "meta-data-gcp",
@@ -780,7 +798,7 @@ var AgentStartCommand = cli.Command{
 			Usage:  "When --disconnect-after-job is specified, the number of seconds to wait for a job before shutting down",
 			EnvVar: "BUILDKITE_AGENT_DISCONNECT_AFTER_JOB_TIMEOUT",
 		},
-	},
+	),
 	Action: func(c *cli.Context) error {
 		ctx := context.Background()
 		ctx, cfg, l, configFile, done := setupLoggerAndConfig[AgentStartConfig](ctx, c, withConfigFilePaths(
@@ -790,18 +808,7 @@ var AgentStartCommand = cli.Command{
 
 		// used later to force the shutdown of the agent
 		ctx, cancel := context.WithCancel(ctx)
-
-		// Verify that the bootstrap buildkite-agent executable matches the current host agent
-		// executable. In development builds, it exits on mismatch; otherwise it only logs a warning.
-		// This is to avoid confusion when making changes to the buildkite-agent but forgetting to
-		// update the buildkite-agent binary available from $PATH
-		if err := checkBinaryPaths(); err != nil {
-			if version.IsDevelopmentBuild() {
-				return fmt.Errorf("check binary paths: %s", err)
-			} else {
-				l.Warn("check binary paths: %s", err)
-			}
-		}
+		defer cancel()
 
 		// Remove any config env from the environment to prevent them propagating to bootstrap
 		if err := UnsetConfigFromEnvironment(c); err != nil {
@@ -1022,6 +1029,7 @@ var AgentStartCommand = cli.Command{
 			CommandEval:                  !cfg.NoCommandEval,
 			PluginsEnabled:               !cfg.NoPlugins,
 			PluginValidation:             !cfg.NoPluginValidation,
+			PluginsAlwaysCloneFresh:      cfg.PluginsAlwaysCloneFresh,
 			LocalHooksEnabled:            !cfg.NoLocalHooks,
 			AllowedEnvironmentVariables:  allowedEnvironmentVariables,
 			StrictSingleHooks:            cfg.StrictSingleHooks,
@@ -1029,7 +1037,8 @@ var AgentStartCommand = cli.Command{
 			ANSITimestamps:               !cfg.NoANSITimestamps,
 			TimestampLines:               cfg.TimestampLines,
 			DisconnectAfterJob:           cfg.DisconnectAfterJob,
-			DisconnectAfterIdleTimeout:   cfg.DisconnectAfterIdleTimeout,
+			DisconnectAfterIdleTimeout:   time.Duration(cfg.DisconnectAfterIdleTimeout) * time.Second,
+			DisconnectAfterUptime:        time.Duration(cfg.DisconnectAfterUptime) * time.Second,
 			CancelGracePeriod:            cfg.CancelGracePeriod,
 			SignalGracePeriod:            signalGracePeriod,
 			EnableJobLogTmpfile:          cfg.EnableJobLogTmpfile,
@@ -1041,6 +1050,7 @@ var AgentStartCommand = cli.Command{
 			AcquireJob:                   cfg.AcquireJob,
 			TracingBackend:               cfg.TracingBackend,
 			TracingServiceName:           cfg.TracingServiceName,
+			TracingPropagateTraceparent:  cfg.TracingPropagateTraceparent,
 			TraceContextEncoding:         cfg.TraceContextEncoding,
 			AllowMultipartArtifactUpload: !cfg.NoMultipartArtifactUpload,
 			KubernetesExec:               cfg.KubernetesExec,
@@ -1061,16 +1071,15 @@ var AgentStartCommand = cli.Command{
 		}
 
 		if cfg.LogFormat == "text" {
-			welcomeMessage :=
-				"\n" +
-					"%s   _           _ _     _ _    _ _                                _\n" +
-					"  | |         (_) |   | | |  (_) |                              | |\n" +
-					"  | |__  _   _ _| | __| | | ___| |_ ___    __ _  __ _  ___ _ __ | |_\n" +
-					"  | '_ \\| | | | | |/ _` | |/ / | __/ _ \\  / _` |/ _` |/ _ \\ '_ \\| __|\n" +
-					"  | |_) | |_| | | | (_| |   <| | ||  __/ | (_| | (_| |  __/ | | | |_\n" +
-					"  |_.__/ \\__,_|_|_|\\__,_|_|\\_\\_|\\__\\___|  \\__,_|\\__, |\\___|_| |_|\\__|\n" +
-					"                                                 __/ |\n" +
-					" https://buildkite.com/agent                    |___/\n%s\n"
+			welcomeMessage := "\n" +
+				"%s   _           _ _     _ _    _ _                                _\n" +
+				"  | |         (_) |   | | |  (_) |                              | |\n" +
+				"  | |__  _   _ _| | __| | | ___| |_ ___    __ _  __ _  ___ _ __ | |_\n" +
+				"  | '_ \\| | | | | |/ _` | |/ / | __/ _ \\  / _` |/ _` |/ _ \\ '_ \\| __|\n" +
+				"  | |_) | |_| | | | (_| |   <| | ||  __/ | (_| | (_| |  __/ | | | |_\n" +
+				"  |_.__/ \\__,_|_|_|\\__,_|_|\\_\\_|\\__\\___|  \\__,_|\\__, |\\___|_| |_|\\__|\n" +
+				"                                                 __/ |\n" +
+				" https://buildkite.com/agent                    |___/\n%s\n"
 
 			if !cfg.NoColor {
 				fmt.Fprintf(os.Stderr, welcomeMessage, "\x1b[38;5;48m", "\x1b[0m")
@@ -1121,7 +1130,11 @@ var AgentStartCommand = cli.Command{
 		}
 
 		if agentConf.DisconnectAfterIdleTimeout > 0 {
-			l.Info("Agents will disconnect after %d seconds of inactivity", agentConf.DisconnectAfterIdleTimeout)
+			l.Info("Agents will disconnect after %v of inactivity", agentConf.DisconnectAfterIdleTimeout)
+		}
+
+		if agentConf.DisconnectAfterUptime > 0 {
+			l.Info("Agents will disconnect after %v of uptime and shut down after any running jobs complete", agentConf.DisconnectAfterUptime)
 		}
 
 		if len(cfg.AllowedRepositories) > 0 {
@@ -1185,7 +1198,7 @@ var AgentStartCommand = cli.Command{
 		// so we may as well check that'll work now and fail early if it's a problem
 		if !osutil.FileExists(agentConf.BuildPath) {
 			l.Info("Build Path doesn't exist, creating it (%s)", agentConf.BuildPath)
-			// Actual file permissions will be reduced by umask, and won't be 0777 unless the user has manually changed the umask to 000
+			// Actual file permissions will be reduced by umask, and won't be 0o777 unless the user has manually changed the umask to 000
 			if err := os.MkdirAll(agentConf.BuildPath, 0o777); err != nil {
 				return fmt.Errorf("failed to create builds path: %w", err)
 			}
@@ -1245,15 +1258,15 @@ var AgentStartCommand = cli.Command{
 			}
 
 			// Register the agent with the buildkite API
-			ag, err := client.Register(ctx, registerReq)
+			reg, err := client.Register(ctx, registerReq)
 			if err != nil {
 				return err
 			}
 
 			// Create an agent worker to run the agent
 			workers = append(workers, agent.NewAgentWorker(
-				l.WithFields(logger.StringField("agent", ag.Name)),
-				ag,
+				l.WithFields(logger.StringField("agent", reg.Name)),
+				reg,
 				mc,
 				apiClient,
 				agent.AgentWorkerConfig{
@@ -1280,7 +1293,15 @@ var AgentStartCommand = cli.Command{
 		}
 
 		// Handle process signals
-		signals := handlePoolSignals(ctx, l, pool, cancel, cfg.CancelGracePeriod)
+		poolSigs := &poolSignals{
+			log:               l,
+			pool:              pool,
+			cancelGracePeriod: time.Duration(cfg.CancelGracePeriod) * time.Second,
+			// Under Kubernetes, there is no user interactively signalling us,
+			// so on SIGTERM, stop un-gracefully.
+			skipGraceful: cfg.KubernetesExec,
+		}
+		signals := poolSigs.handle(ctx)
 		defer close(signals)
 
 		l.Info("Starting %d Agent(s)", cfg.Spawn)
@@ -1291,52 +1312,38 @@ var AgentStartCommand = cli.Command{
 			pool.StartStatusServer(ctx, l, cfg.HealthCheckAddr)
 		}
 
-		err = pool.Start(ctx)
-		if errors.Is(err, core.ErrJobAcquisitionRejected) {
+		var exit core.ProcessExit
+		switch err := pool.Start(ctx); {
+		case errors.Is(err, core.ErrJobAcquisitionRejected):
 			// If the agent tried to acquire a job, but it couldn't because the job was already taken, we should exit with a
 			// specific exit code so that the caller can know that this job can't be acquired.
 
 			const acquisitionFailedExitCode = 27 // chosen by fair dice roll
 			return cli.NewExitError(err, acquisitionFailedExitCode)
+
+		case errors.Is(err, core.ErrJobLocked):
+			// If the agent tried to acquire a job, but it couldn't because the job is locked (waiting for dependencies),
+			// we should exit with a specific exit code so that the caller can know that this job is locked.
+
+			const jobLockedExitCode = 28
+			return cli.NewExitError(err, jobLockedExitCode)
+
+		case errors.As(err, &exit):
+			if cfg.ReflectExitStatus {
+				// If the agent acquired a job and it failed or was cancelled,
+				// then report its exit code as our own.
+				return cli.NewExitError(err, exit.Status)
+			}
+			// The job ran. Even though it failed, the agent did its job.
+			return nil
+
+		default:
+			return err
+
 		}
-
-		return err
 	},
 }
 
-// checkBinaryPaths looks up both the bootstrap and host buildkite-agent paths,
-// and returns an error if they do not match or if either cannot be determined.
-func checkBinaryPaths() error {
-	bootstrapPath, err := exec.LookPath("buildkite-agent")
-	if err != nil {
-		return fmt.Errorf("failed to locate bootstrap buildkite-agent: %w", err)
-	}
-
-	evalBootstrapPath, err := filepath.EvalSymlinks(bootstrapPath)
-	if err != nil {
-		return fmt.Errorf("failed to locate bootstrap buildkite-agent: %w", err)
-	}
-
-	hostPath, err := os.Executable()
-	if err != nil {
-		return fmt.Errorf("failed to determine host buildkite-agent executable: %w", err)
-	}
-
-	evalHostPath, err := filepath.EvalSymlinks(hostPath)
-	if err != nil {
-		return fmt.Errorf("failed to determine host buildkite-agent executable: %w", err)
-	}
-
-	if evalHostPath != evalBootstrapPath {
-		return fmt.Errorf(
-			"mismatched buildkite-agent paths: host=%q bootstrap=%q",
-			evalHostPath, evalBootstrapPath,
-		)
-	}
-
-	return nil
-}
-
 func parseAndValidateJWKS(ctx context.Context, keysetType, path string) (jwk.Set, error) {
 	jwksBytes, err := os.ReadFile(path)
 	if err != nil {
@@ -1368,70 +1375,86 @@ func parseAndValidateJWKS(ctx context.Context, keysetType, path string) (jwk.Set
 	return jwks, nil
 }
 
-func handlePoolSignals(ctx context.Context, l logger.Logger, pool *agent.AgentPool, cancel context.CancelFunc, cancelGracePeriod int) chan os.Signal {
+type poolSignals struct {
+	log               logger.Logger
+	pool              *agent.AgentPool
+	cancelGracePeriod time.Duration
+	skipGraceful      bool
+}
+
+func (ps *poolSignals) handle(ctx context.Context) chan os.Signal {
 	signals := make(chan os.Signal, 1)
-	signal.Notify(signals, os.Interrupt,
+	signal.Notify(
+		signals,
+		os.Interrupt,
 		syscall.SIGHUP,
 		syscall.SIGTERM,
 		syscall.SIGINT,
-		syscall.SIGQUIT)
-
-	go func() {
-		_, setStatus, done := status.AddSimpleItem(ctx, "Handle Pool Signals")
-		defer done()
-		setStatus("⏳ Waiting for a signal")
-
-		var interruptCount int
-
-		for sig := range signals {
-			l.Debug("Received signal `%v`", sig)
-			setStatus(fmt.Sprintf("Received signal `%v`", sig))
-
-			switch sig {
-			case syscall.SIGQUIT:
-				l.Debug("Received signal `%s`", sig.String())
-				pool.Stop(false)
-			case syscall.SIGTERM, syscall.SIGINT:
-				l.Debug("Received signal `%s`", sig.String())
-				if interruptCount == 0 {
-					interruptCount++
-					l.Info("Received CTRL-C, send again to forcefully kill the agent(s)")
-					pool.Stop(true)
-				} else {
-					l.Info("Forcefully stopping running jobs and stopping the agent(s) in %d seconds", cancelGracePeriod)
-
-					gracefulContext, _ := job.WithGracePeriod(ctx, time.Duration(max(cancelGracePeriod, minGracePeriod))*time.Second)
-
-					go func() {
-						l.Info("Forced agent(s) to stop")
-						pool.Stop(false) // one last chance to stop
-
-						// Wait half the grace period before cancelling the context
-						time.Sleep(time.Duration(max(cancelGracePeriod/2, minGracePeriod/2)) * time.Second)
-
-						l.Info("Cancelling all internal tasks and API requests")
-						cancel() // cancel the context to stop all network operations
-					}()
-
-					// Once pending retries and requests are cancelled,
-					// the main goroutine should exit before this grace period expires, ending the program.
-					// If that doesn't happen, exit 1 below.
-					<-gracefulContext.Done()
-					l.Info("exiting with status 1")
-
-					// this should only be called if the context cancellation and forceful stop fails
-					os.Exit(1)
-
-				}
-			default:
-				l.Debug("Ignoring signal `%s`", sig.String())
-			}
-		}
-	}()
+		syscall.SIGQUIT,
+	)
 
+	go ps.handleLoop(ctx, signals)
 	return signals
 }
 
+func (ps *poolSignals) handleLoop(ctx context.Context, signals chan os.Signal) {
+	_, setStatus, done := status.AddSimpleItem(ctx, "Handle Pool Signals")
+	defer done()
+	setStatus("⏳ Waiting for a signal")
+
+	interruptCount := 0
+	if ps.skipGraceful {
+		interruptCount = 1
+	}
+
+	ungracefulStop := func() {
+		// We shouldn't block the signal handler loop either by waiting
+		// for the jobs to cancel or by waiting for the cancel grace
+		// period to expire.
+		go ps.pool.StopUngracefully() // one last chance to stop
+		go func() {
+			// Assuming cancelling jobs takes the full cancel grace period,
+			// allow 1 second to send agent disconnects.
+			time.Sleep(ps.cancelGracePeriod + 1*time.Second)
+			// We get here if the main goroutine hasn't returned yet.
+			ps.log.Info("Timed out waiting for agents to exit; exiting immediately with status 1")
+			os.Exit(1)
+		}()
+	}
+
+	for sig := range signals {
+		ps.log.Debug("Received signal `%v`", sig)
+		setStatus(fmt.Sprintf("Received signal `%v`", sig))
+
+		switch sig {
+		case syscall.SIGQUIT:
+			ungracefulStop()
+
+		case syscall.SIGTERM, syscall.SIGINT:
+			interruptCount++
+			switch interruptCount {
+			case 1:
+				ps.log.Info("Received CTRL-C, send again to forcefully kill the agent(s)")
+				ps.pool.StopGracefully()
+
+			case 2:
+				ps.log.Info("Forcefully stopping running jobs and stopping the agent(s) in %v", ps.cancelGracePeriod)
+				if !ps.skipGraceful {
+					ps.log.Info("Press Ctrl-C one more time to exit immediately without disconnecting - note that agents will be considered lost!")
+				}
+				ungracefulStop()
+
+			case 3:
+				ps.log.Info("Exiting immediately with status 1")
+				os.Exit(1)
+			}
+
+		default:
+			ps.log.Debug("Ignoring signal `%s`", sig.String())
+		}
+	}
+}
+
 func agentStartupHook(log logger.Logger, cfg AgentStartConfig) error {
 	return agentLifecycleHook("agent-startup", log, cfg)
 }
@@ -1446,7 +1469,7 @@ func agentShutdownHook(log logger.Logger, cfg AgentStartConfig) {
 func agentLifecycleHook(hookName string, log logger.Logger, cfg AgentStartConfig) error {
 	// search for hook (including .bat & .ps1 files on Windows)
 	hooks := []string{}
-	p, err := hook.Find(cfg.HooksPath, hookName)
+	p, err := hook.Find(nil, cfg.HooksPath, hookName)
 	if err != nil {
 		if !os.IsNotExist(err) {
 			log.Error("Error finding %q hook: %v", hookName, err)
@@ -1458,7 +1481,7 @@ func agentLifecycleHook(hookName string, log logger.Logger, cfg AgentStartConfig
 
 	// also search for hook in any additionally provided locations
 	for _, h := range cfg.AdditionalHooksPaths {
-		p, err = hook.Find(h, hookName)
+		p, err = hook.Find(nil, h, hookName)
 		if err != nil {
 			if !os.IsNotExist(err) {
 				log.Error("Error finding %q hook: %v", hookName, err)
@@ -1592,5 +1615,4 @@ func leaderPinger(ctx context.Context, l logger.Logger, path, leaderPath string)
 func envHasKey(key string) bool {
 	_, ok := os.LookupEnv(key)
 	return ok
-
 }
