@@ -27,8 +27,10 @@ func TestAgentStop(t *testing.T) {
 
 	ctx := context.Background()
 	cfg := AgentStopConfig{
-		AgentAccessToken: "agentaccesstoken",
-		Endpoint:         server.URL,
+		APIConfig: APIConfig{
+			AgentAccessToken: "agentaccesstoken",
+			Endpoint:         server.URL,
+		},
 	}
 	l := logger.NewBuffer()
 
