@@ -43,6 +43,16 @@ var BuildkiteAgentCommands = []cli.Command{
 			BuildCancelCommand,
 		},
 	},
+	{
+		Name:     "cache",
+		Category: categoryJobCommands,
+		Usage:    "Manage build caches",
+		Hidden:   true, // currently in experimental phase
+		Subcommands: []cli.Command{
+			CacheSaveCommand,
+			CacheRestoreCommand,
+		},
+	},
 	{
 		Name:     "env",
 		Category: categoryJobCommands,
