@@ -70,7 +70,7 @@ Example:
     $ buildkite-agent pipeline upload my-custom-pipeline.yml
     $ ./script/dynamic_step_generator | buildkite-agent pipeline upload`
 
-const ifChangedSkippedMsg = "if_changed pattern did not match any paths changed in this build"
+const ifChangedSkippedMsg = "if_changed matched no unexcluded changed files in this build"
 
 const defaultGitDiffBase = "origin/main"
 
@@ -137,8 +137,8 @@ var PipelineUploadCommand = cli.Command{
 		},
 		cli.BoolTFlag{
 			Name:   "apply-if-changed",
-			Usage:  "When enabled, steps containing an ′if_changed′ key are evaluated against the git diff. If the ′if_changed′ glob pattern match no files changed in the build, the step is skipped.",
-			EnvVar: "BUILDKITE_AGENT_APPLY_SKIP_IF_UNCHANGED",
+			Usage:  "When enabled, steps containing an ′if_changed′ key are evaluated against the git diff. If the ′if_changed′ glob pattern match no files changed in the build, the step is skipped. Minimum Buildkite Agent version: v3.99 (with --apply-if-changed flag), v3.103.0 (enabled by default)",
+			EnvVar: "BUILDKITE_AGENT_APPLY_IF_CHANGED,BUILDKITE_AGENT_APPLY_SKIP_IF_UNCHANGED",
 		},
 		cli.StringFlag{
 			Name:   "git-diff-base",
@@ -319,12 +319,13 @@ var PipelineUploadCommand = cli.Command{
 				if len(cfg.RedactedVars) > 0 {
 					// Secret detection uses the original environment, since
 					// Interpolate merges the pipeline's env block into `environ`.
-					searchForSecrets(l, &cfg, environ, result, input.name)
+					err := searchForSecrets(l, &cfg, environ, result, input.name)
+					if err != nil {
+						return NewExitError(1, err)
+					}
 				}
 
-				var (
-					key signature.Key
-				)
+				var key signature.Key
 
 				switch {
 				case cfg.SigningAWSKMSKey != "":
@@ -509,6 +510,7 @@ func searchForSecrets(
 	if err != nil {
 		l.Warn("Couldn't match environment variable names against redacted-vars: %v", err)
 	}
+
 	for _, name := range short {
 		shortValues[name] = struct{}{}
 	}
@@ -612,25 +614,89 @@ func (cfg *PipelineUploadConfig) parseAndInterpolate(ctx context.Context, src st
 }
 
 // gatherChangedFiles determines changed files in this build.
-func gatherChangedFiles(diffBase string) (mergeBase string, changedPaths []string, err error) {
-	// The --merge-base flag was only added to git-diff recently.
-	mergeBaseOut, err := exec.Command("git", "merge-base", diffBase, "HEAD").Output()
+func gatherChangedFiles(l logger.Logger, diffBase string) (changedPaths []string, err error) {
+	// Corporate needs you to find the differences between diffBase and HEAD.
+	diffBaseCommit, err := exec.Command("git", "rev-parse", diffBase).Output()
 	if err != nil {
-		return "", nil, gitMergeBaseError{diffBase: diffBase, wrapped: err}
+		return nil, gitRevParseError{arg: diffBase, wrapped: err}
 	}
-	mergeBase = strings.TrimSpace(string(mergeBaseOut))
+	headCommit, err := exec.Command("git", "rev-parse", "HEAD").Output()
+	if err != nil {
+		return nil, gitRevParseError{arg: "HEAD", wrapped: err}
+	}
+	if strings.TrimSpace(string(diffBaseCommit)) == strings.TrimSpace(string(headCommit)) {
+		// They're the same commit, so `git diff` will (correctly) report no
+		// changes between them.
+		// This often happens for builds triggered on main, when:
+		// * a PR is merged. HEAD is hopefully either a merge commit or a
+		//   squash commit.
+		// * the developers commit and push directly to main.
+		// As a heuristic, we look at changed files between the latest commit
+		// and its parent (the first parent, if it is a merge commit).
+		// If _multiple_ commits were pushed at once, then this approach will
+		// miss changes from earlier commits. Thus, log a warning.
+		l.Warn("Applying if_changed conditions relative to the first parent of HEAD (because HEAD = %q)", diffBase)
+		l.Warn("If this build is intended to include more than one commit on this branch, if_changed may calculate an incomplete diff. You may need to adjust the --git-diff-base flag or BUILDKITE_GIT_DIFF_BASE env var to choose a different base commit for calculating diffs.")
 
-	gitDiff, err := exec.Command("git", "diff", "--name-only", mergeBase).Output()
-	if err != nil {
-		return mergeBase, nil, gitDiffError{mergeBase: mergeBase, wrapped: err}
+		// Flag Explainer:
+		// `--first-parent`: when reaching a merge commit, follow the first parent
+		// (the branch that was merged *into*). Implies --diff-merges=first-parent,
+		// which is important for turning merge commits into changes.
+		// `-1`: show only the most recent entry from the log
+		// `--name-only`: ensure the names of each changed file are printed
+		// `--pretty='format:'`: only print the names of the files.
+		gitLog, err := exec.Command("git", "log", "--first-parent", "-1", "--name-only", "--pretty=format:").Output()
+		if err != nil {
+			return nil, gitLogError{wrapped: err}
+		}
+		changedPaths = strings.Split(string(gitLog), "\n")
+	} else {
+		// The --merge-base flag was only added to git-diff recently.
+		mergeBaseOut, err := exec.Command("git", "merge-base", diffBase, "HEAD").Output()
+		if err != nil {
+			return nil, gitMergeBaseError{diffBase: diffBase, wrapped: err}
+		}
+		mergeBase := strings.TrimSpace(string(mergeBaseOut))
+		l.Info("Applying if_changed conditions relative to %q (the merge-base of %q and HEAD)", mergeBase, diffBase)
+
+		gitDiff, err := exec.Command("git", "diff", "--name-only", mergeBase).Output()
+		if err != nil {
+			return nil, gitDiffError{mergeBase: mergeBase, wrapped: err}
+		}
+		changedPaths = strings.Split(string(gitDiff), "\n")
 	}
-	changedPaths = strings.Split(string(gitDiff), "\n")
 	changedPaths = slices.DeleteFunc(changedPaths, func(s string) bool {
 		return strings.TrimSpace(s) == ""
 	})
-	return mergeBase, changedPaths, nil
+	plural := "files"
+	if len(changedPaths) == 1 {
+		plural = "file"
+	}
+	l.Info("if_changed found %d changed %s", len(changedPaths), plural)
+	return changedPaths, nil
 }
 
+type gitRevParseError struct {
+	arg     string
+	wrapped error
+}
+
+func (e gitRevParseError) Error() string {
+	return fmt.Sprintf("git rev-parse %q: %v", e.arg, e.wrapped)
+}
+
+func (e gitRevParseError) Unwrap() error { return e.wrapped }
+
+type gitLogError struct {
+	wrapped error
+}
+
+func (e gitLogError) Error() string {
+	return fmt.Sprintf("git log --first-parent -1 --name-only --pretty=\"format:\": %v", e.wrapped)
+}
+
+func (e gitLogError) Unwrap() error { return e.wrapped }
+
 type gitMergeBaseError struct {
 	diffBase string
 	wrapped  error
@@ -684,14 +750,21 @@ stepsLoop:
 			content = st.Contents
 		}
 
-		// Retrieve the value then delete it.
-		// It's not yet understood by the backend.
-		ic := content["if_changed"]
-		if ic == nil {
+		// If content is nil then there's nothing containing an `if_changed`.
+		if content == nil {
 			continue
 		}
+
+		// Retrieve the if_changed value, then delete it.
+		// It is not yet understood by the backend.
+		ifChangedValue := content["if_changed"]
 		delete(content, "if_changed")
 
+		// If there's no if_changed, then the step is unconditional.
+		if ifChangedValue == nil {
+			continue
+		}
+
 		if !ica.enabled {
 			// Not applying the if_changed; leaving it deleted.
 			continue
@@ -699,10 +772,21 @@ stepsLoop:
 
 		// If we don't know the changed paths yet, call out to Git.
 		if !ica.gathered {
-			mergeBase, cps, err := gatherChangedFiles(ica.diffBase)
+			cps, err := gatherChangedFiles(l, ica.diffBase)
 			if err != nil {
 				l.Error("Couldn't determine git diff from upstream, not skipping any pipeline steps: %v", err)
+				var exitErr *exec.ExitError
+				if errors.As(err, &exitErr) && len(exitErr.Stderr) > 0 {
+					// stderr came from git, which is typically human readable
+					l.Error("git: %s", exitErr.Stderr)
+				}
 				switch err := err.(type) {
+				case gitRevParseError:
+					l.Error("This could be because %q might not be a commit in the repository.\n"+
+						"You may need to change the --git-diff-base flag or BUILDKITE_GIT_DIFF_BASE env var.",
+						err.arg,
+					)
+
 				case gitMergeBaseError:
 					l.Error("This could be because %q might not be a commit in the repository.\n"+
 						"You may need to change the --git-diff-base flag or BUILDKITE_GIT_DIFF_BASE env var.",
@@ -723,33 +807,109 @@ stepsLoop:
 			}
 
 			// The changed files are now known.
-			l.Info("Applying if_changed conditions relative to %q (the merge-base of %q and HEAD) - %d changed files", mergeBase, ica.diffBase, len(cps))
 			ica.gathered = true
 			ica.changedPaths = cps
 		}
 
-		// Parse and test the glob pattern against the paths.
-		pattern, ok := ic.(string)
-		if !ok {
-			l.Warn("if_changed value must be a string containing a glob pattern (was a %T)\n"+
-				"The step will not be skipped.", ic)
-			continue
+		var include, exclude []*zzglob.Pattern
+		switch x := ifChangedValue.(type) {
+		case *ordered.MapSA:
+			// Object form:
+			// if_changed:
+			//   include: (required; string or list)
+			//   exclude: (optional; string or list)
+			inclVal, has := x.Get("include")
+			if !has {
+				l.Warn("The value for if_changed was a mapping, but it didn't have an `include` key. The step will not be skipped.")
+				continue stepsLoop
+			}
+			var err error
+			include, err = ifChangedPatterns(inclVal)
+			if err != nil {
+				l.Warn("Couldn't parse if_changed.include patterns: %v. The step will not be skipped.", err)
+				continue stepsLoop
+			}
+			exclVal, has := x.Get("exclude")
+			if !has {
+				break // switch
+			}
+			exclude, err = ifChangedPatterns(exclVal)
+			if err != nil {
+				l.Warn("Couldn't parse if_changed.exclude patterns: %v. The step will not be skipped.", err)
+				continue stepsLoop
+			}
+
+		default:
+			// Should be either a simple string or a list of strings.
+			inc, err := ifChangedPatterns(x)
+			if err != nil {
+				l.Warn("Couldn't parse if_changed patterns: %v. The step will not be skipped.", err)
+				continue stepsLoop
+			}
+			include = inc
 		}
 
-		glob, err := zzglob.Parse(pattern)
-		if err != nil {
-			l.Warn("if_changed value %q couldn't be parsed as a glob pattern: %v\n"+
-				"The step will not be skipped.", pattern, err)
-			continue
-		}
+		// For each remaining changed path, test it against each pattern in
+		// `exclude` and `include`.
+		// If it's included, and wasn't excluded, we _don't skip_ this step.
+		// Got it?
+	pathsLoop:
 		for _, cp := range ica.changedPaths {
-			if glob.Match(cp) {
-				continue stepsLoop // one of the paths changed
+			// First, check if the path is excluded.
+			for _, g := range exclude {
+				if g.Match(cp) {
+					continue pathsLoop
+				}
+			}
+			// It's not excluded! now check if it is included.
+			for _, g := range include {
+				if g.Match(cp) {
+					continue stepsLoop
+				}
 			}
 		}
 
-		// Glob matched no changed file paths - this step is now skipped.
+		// The globs matched no changed file paths - this step is now skipped.
 		// Note that the "skip" string is limited to 70 characters.
 		content["skip"] = ifChangedSkippedMsg
 	}
 }
+
+// ifChangedPatterns converts a string or list within `if_changed` into a slice
+// of parsed globs.
+func ifChangedPatterns(value any) ([]*zzglob.Pattern, error) {
+	if value == nil {
+		return nil, nil
+	}
+
+	var patterns []string
+	switch x := value.(type) {
+	case string:
+		// A single string.
+		patterns = []string{x}
+
+	case []any:
+		// A list of strings.
+		for i, item := range x {
+			patt, ok := item.(string)
+			if !ok {
+				return nil, fmt.Errorf("at index %d, the item had unsupported type %T", i, item)
+			}
+			patterns = append(patterns, patt)
+		}
+
+	default:
+		return nil, fmt.Errorf("the value had unsupported type %T", x)
+	}
+
+	globs := make([]*zzglob.Pattern, 0, len(patterns))
+	for _, patt := range patterns {
+		g, err := zzglob.Parse(patt)
+		if err != nil {
+			return nil, fmt.Errorf("while parsing glob pattern %q: %w", patt, err)
+		}
+		globs = append(globs, g)
+	}
+
+	return globs, nil
+}
