@@ -343,12 +343,27 @@ func TestIfChangedApplicator(t *testing.T) {
 					"if_changed": "foo/**",
 				},
 			},
+			&pipeline.CommandStep{
+				Command: "only runs when files in foo changed, except for baz",
+				RemainingFields: map[string]any{
+					"if_changed": ordered.MapFromItems(
+						ordered.TupleSA{Key: "include", Value: "foo/**"},
+						ordered.TupleSA{Key: "exclude", Value: "foo/baz"},
+					),
+				},
+			},
 			&pipeline.CommandStep{
 				Command: "only runs when files in bar changed",
 				RemainingFields: map[string]any{
 					"if_changed": "bar/**",
 				},
 			},
+			&pipeline.CommandStep{
+				Command: "only runs when files in foo or bar changed",
+				RemainingFields: map[string]any{
+					"if_changed": []any{"foo/**", "bar/**"},
+				},
+			},
 			&pipeline.CommandStep{
 				Command: "only runs when any files changed",
 				RemainingFields: map[string]any{
@@ -374,10 +389,18 @@ func TestIfChangedApplicator(t *testing.T) {
 					Command:         "only runs when files in foo changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when files in bar changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when any files changed",
 					RemainingFields: map[string]any{},
@@ -399,12 +422,24 @@ func TestIfChangedApplicator(t *testing.T) {
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
 				&pipeline.CommandStep{
 					Command: "only runs when files in bar changed",
 					RemainingFields: map[string]any{
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
 				&pipeline.CommandStep{
 					Command: "only runs when any files changed",
 					RemainingFields: map[string]any{
@@ -414,7 +449,7 @@ func TestIfChangedApplicator(t *testing.T) {
 			},
 		},
 		{
-			name: "root file change",
+			name: "change in qux",
 			ica: &ifChangedApplicator{
 				enabled:      true,
 				gathered:     true, // pretend we ran git diff
@@ -428,12 +463,24 @@ func TestIfChangedApplicator(t *testing.T) {
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
 				&pipeline.CommandStep{
 					Command: "only runs when files in bar changed",
 					RemainingFields: map[string]any{
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when any files changed",
 					RemainingFields: map[string]any{},
@@ -441,7 +488,7 @@ func TestIfChangedApplicator(t *testing.T) {
 			},
 		},
 		{
-			name: "change in foo",
+			name: "change in foo/README.md",
 			ica: &ifChangedApplicator{
 				enabled:      true,
 				gathered:     true, // pretend we ran git diff
@@ -453,12 +500,56 @@ func TestIfChangedApplicator(t *testing.T) {
 					Command:         "only runs when files in foo changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command: "only runs when files in bar changed",
 					RemainingFields: map[string]any{
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{},
+				},
+				&pipeline.CommandStep{
+					Command:         "only runs when any files changed",
+					RemainingFields: map[string]any{},
+				},
+			},
+		},
+
+		{
+			name: "change in foo/baz",
+			ica: &ifChangedApplicator{
+				enabled:      true,
+				gathered:     true, // pretend we ran git diff
+				changedPaths: []string{"foo/baz"},
+			},
+			want: pipeline.Steps{
+				&pipeline.CommandStep{Command: "always runs"},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo changed",
+					RemainingFields: map[string]any{},
+				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in bar changed",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when any files changed",
 					RemainingFields: map[string]any{},
@@ -466,7 +557,7 @@ func TestIfChangedApplicator(t *testing.T) {
 			},
 		},
 		{
-			name: "change in bar",
+			name: "change in bar/README.md",
 			ica: &ifChangedApplicator{
 				enabled:      true,
 				gathered:     true, // pretend we ran git diff
@@ -480,10 +571,20 @@ func TestIfChangedApplicator(t *testing.T) {
 						"skip": ifChangedSkippedMsg,
 					},
 				},
+				&pipeline.CommandStep{
+					Command: "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{
+						"skip": ifChangedSkippedMsg,
+					},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when files in bar changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when any files changed",
 					RemainingFields: map[string]any{},
@@ -491,11 +592,11 @@ func TestIfChangedApplicator(t *testing.T) {
 			},
 		},
 		{
-			name: "change in foo and bar",
+			name: "changes in foo/hello.go, foo/baz, and bar/README.md",
 			ica: &ifChangedApplicator{
 				enabled:      true,
 				gathered:     true, // pretend we ran git diff
-				changedPaths: []string{"foo/hello.go", "bar/README.md"},
+				changedPaths: []string{"foo/hello.go", "foo/baz", "bar/README.md"},
 			},
 			want: pipeline.Steps{
 				&pipeline.CommandStep{Command: "always runs"},
@@ -503,10 +604,19 @@ func TestIfChangedApplicator(t *testing.T) {
 					Command:         "only runs when files in foo changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					// A file other than baz changed in foo, so this runs.
+					Command:         "only runs when files in foo changed, except for baz",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when files in bar changed",
 					RemainingFields: map[string]any{},
 				},
+				&pipeline.CommandStep{
+					Command:         "only runs when files in foo or bar changed",
+					RemainingFields: map[string]any{},
+				},
 				&pipeline.CommandStep{
 					Command:         "only runs when any files changed",
 					RemainingFields: map[string]any{},
@@ -535,7 +645,7 @@ func TestIfChangedApplicator_WeirdPipeline(t *testing.T) {
 
 	steps := pipeline.Steps{
 		&pipeline.CommandStep{
-			Command: "wrong type for if_changed",
+			Command: "unsupported type for if_changed",
 			RemainingFields: map[string]any{
 				"if_changed": 42,
 			},
@@ -546,6 +656,23 @@ func TestIfChangedApplicator_WeirdPipeline(t *testing.T) {
 				"if_changed": "bar/**/[asdf[[[[asdf",
 			},
 		},
+		&pipeline.CommandStep{
+			Command: "invalid exclude pattern",
+			RemainingFields: map[string]any{
+				"if_changed": ordered.MapFromItems(
+					ordered.TupleSA{Key: "include", Value: "**"},
+					ordered.TupleSA{Key: "exclude", Value: "{a{b{c{d"},
+				),
+			},
+		},
+		&pipeline.CommandStep{
+			Command: "mapping without include",
+			RemainingFields: map[string]any{
+				"if_changed": ordered.MapFromItems(
+					ordered.TupleSA{Key: "exclude", Value: "asdf"},
+				),
+			},
+		},
 		&pipeline.TriggerStep{
 			Contents: map[string]any{
 				"if_changed": "version/VERSION",
@@ -568,13 +695,21 @@ func TestIfChangedApplicator_WeirdPipeline(t *testing.T) {
 
 	want := pipeline.Steps{
 		&pipeline.CommandStep{
-			Command:         "wrong type for if_changed",
+			Command:         "unsupported type for if_changed",
 			RemainingFields: map[string]any{},
 		},
 		&pipeline.CommandStep{
 			Command:         "invalid glob pattern",
 			RemainingFields: map[string]any{},
 		},
+		&pipeline.CommandStep{
+			Command:         "invalid exclude pattern",
+			RemainingFields: map[string]any{},
+		},
+		&pipeline.CommandStep{
+			Command:         "mapping without include",
+			RemainingFields: map[string]any{},
+		},
 		&pipeline.TriggerStep{
 			Contents: map[string]any{
 				"skip": ifChangedSkippedMsg,
@@ -605,5 +740,4 @@ func TestIfChangedApplicator_WeirdPipeline(t *testing.T) {
 	if diff := cmp.Diff(steps, want); diff != "" {
 		t.Errorf("after ica.apply(l, steps) (-got, +want):\n%s", diff)
 	}
-
 }
