@@ -8,9 +8,8 @@ import (
 )
 
 func BenchmarkBMRedactor(b *testing.B) {
-	b.ResetTimer()
 	r := NewBMRedactor(io.Discard, "[REDACTED]", bigLipsumSecrets)
-	for range b.N {
+	for b.Loop() {
 		if _, err := fmt.Fprintln(r, bigLipsum); err != nil {
 			b.Errorf("fmt.Fprintln(r, bigLipsum) error = %v", err)
 		}
@@ -116,7 +115,6 @@ func (redactor *BoyerMooreRedactor) Reset(needles []string) {
 			}
 		}
 	}
-
 }
 
 func (redactor *BoyerMooreRedactor) Write(input []byte) (int, error) {
