@@ -107,7 +107,6 @@ func TestLoggerStreamer(t *testing.T) {
 
 	//nolint:errcheck // Writes to bytes.Buffer never error.
 	func() {
-
 		fmt.Fprintln(want, "TEST># Rest of the line")
 		fmt.Fprintln(want, "TEST># And another")
 		fmt.Fprintln(want, "TEST># No line end")
@@ -124,7 +123,7 @@ func BenchmarkDoubleFmt(b *testing.B) {
 		fmt.Fprintf(io.Discard, "%s", fmt.Sprintf(format, v...))
 		fmt.Fprintln(io.Discard)
 	}
-	for range b.N {
+	for b.Loop() {
 		logf("asdfghjkl %s %d %t", "hi", 42, true)
 	}
 }
@@ -134,7 +133,7 @@ func BenchmarkFmtConcat(b *testing.B) {
 	logf := func(format string, v ...any) {
 		fmt.Fprintf(io.Discard, format+"\n", v...)
 	}
-	for range b.N {
+	for b.Loop() {
 		logf("asdfghjkl %s %d %t", "hi", 42, true)
 	}
 }
