@@ -1,3 +1,6 @@
+// SPDX-FileCopyrightText: Copyright 2015-2025 go-swagger maintainers
+// SPDX-License-Identifier: Apache-2.0
+
 package loading
 
 import (
@@ -5,8 +8,9 @@ import (
 	"net/http/httptest"
 	"testing"
 
-	"github.com/stretchr/testify/assert"
-	"github.com/stretchr/testify/require"
+	_ "github.com/go-openapi/testify/enable/yaml/v2" // enable YAMLEq in testify
+	"github.com/go-openapi/testify/v2/assert"
+	"github.com/go-openapi/testify/v2/require"
 )
 
 func TestYAMLMatcher(t *testing.T) {
@@ -19,7 +23,7 @@ func TestYAMLMatcher(t *testing.T) {
 
 func TestYAMLDoc(t *testing.T) {
 	t.Run("should retrieve pet store API as YAML", func(t *testing.T) {
-		serv := httptest.NewServer(http.HandlerFunc(serveYAMLPestore))
+		serv := httptest.NewServer(http.HandlerFunc(serveYAMLPetStore))
 		defer serv.Close()
 
 		s, err := YAMLDoc(serv.URL)
