@@ -16,18 +16,57 @@
 package config
 
 import (
+	"bytes"
 	"context"
+	"crypto/rand"
+	"crypto/rsa"
+	"crypto/tls"
+	"crypto/x509"
+	"crypto/x509/pkix"
+	"encoding/json"
+	"encoding/pem"
 	"fmt"
+	"math/big"
+	"net"
+	"net/http"
+	"net/http/httptest"
 	"net/url"
+	"os"
 	"reflect"
+	"strings"
+	"sync/atomic"
 	"testing"
+	"time"
 
 	"github.com/coreos/go-oidc/v3/oidc"
-	lru "github.com/hashicorp/golang-lru"
+	lru "github.com/hashicorp/golang-lru/v2"
 	"github.com/sigstore/fulcio/pkg/generated/protobuf"
 )
 
-var validCfg = `
+var validYamlCfg = `
+oidc-issuers:
+  https://accounts.google.com:
+    issuer-url: https://accounts.google.com
+    client-id: foo
+    type: email
+    challenge-claim: email
+  https://internal.example.com:
+    issuer-url: https://internal.example.com
+    client-id: foo
+    type: email
+    challenge-claim: email
+    skip-email-verification: true
+meta-issuers:
+  https://oidc.eks.*.amazonaws.com/id/*:
+    client-id: bar
+    type: kubernetes
+  https://oidc.foo.*.bar.com/id/*:
+    client-id: bar
+    type: ci-provider
+    ci-provider: github-workflow
+`
+
+var validJSONCfg = `
 {
 	"OIDCIssuers": {
 		"https://accounts.google.com": {
@@ -35,12 +74,24 @@ var validCfg = `
 			"ClientID": "foo",
 			"Type": "email",
 			"ChallengeClaim": "email"
+		},
+		"https://internal.example.com": {
+			"IssuerURL": "https://internal.example.com",
+			"ClientID": "foo",
+			"Type": "email",
+			"ChallengeClaim": "email",
+			"SkipEmailVerification": true
 		}
 	},
 	"MetaIssuers": {
 		"https://oidc.eks.*.amazonaws.com/id/*": {
 			"ClientID": "bar",
 			"Type": "kubernetes"
+		},
+		"https://oidc.foo.*.bar.com/id/*": {
+			"ClientID": "bar",
+			"Type": "ci-provider",
+			"CiProvider": "github-workflow"
 		}
 	}
 }
@@ -473,6 +524,42 @@ func Test_validateAllowedDomain(t *testing.T) {
 	}
 }
 
+func TestSkipEmailVerificationParsing(t *testing.T) {
+	yamlConfig := `
+oidc-issuers:
+  https://internal.example.com:
+    issuer-url: https://internal.example.com
+    client-id: sigstore
+    type: email
+    skip-email-verification: true
+  https://public.example.com:
+    issuer-url: https://public.example.com
+    client-id: sigstore
+    type: email
+`
+
+	cfg, err := parseConfig([]byte(yamlConfig))
+	if err != nil {
+		t.Fatalf("failed to parse config: %v", err)
+	}
+
+	internalIssuer, ok := cfg.OIDCIssuers["https://internal.example.com"]
+	if !ok {
+		t.Fatal("internal issuer not found")
+	}
+	if !internalIssuer.SkipEmailVerification {
+		t.Error("expected SkipEmailVerification to be true for internal issuer")
+	}
+
+	publicIssuer, ok := cfg.OIDCIssuers["https://public.example.com"]
+	if !ok {
+		t.Fatal("public issuer not found")
+	}
+	if publicIssuer.SkipEmailVerification {
+		t.Error("expected SkipEmailVerification to be false (default) for public issuer")
+	}
+}
+
 func Test_issuerToChallengeClaim(t *testing.T) {
 	if claim := issuerToChallengeClaim(IssuerTypeEmail, ""); claim != "email" {
 		t.Fatalf("expected email subject claim for email issuer, got %s", claim)
@@ -492,6 +579,9 @@ func Test_issuerToChallengeClaim(t *testing.T) {
 	if claim := issuerToChallengeClaim(IssuerTypeGithubWorkflow, ""); claim != "sub" {
 		t.Fatalf("expected sub subject claim for GitHub issuer, got %s", claim)
 	}
+	if claim := issuerToChallengeClaim(IssuerTypeCIProvider, ""); claim != "sub" {
+		t.Fatalf("expected sub subject claim for CI issuer, got %s", claim)
+	}
 	if claim := issuerToChallengeClaim(IssuerTypeGitLabPipeline, ""); claim != "sub" {
 		t.Fatalf("expected sub subject claim for GitLab issuer, got %s", claim)
 	}
@@ -515,51 +605,103 @@ func Test_issuerToChallengeClaim(t *testing.T) {
 }
 
 func TestToIssuers(t *testing.T) {
-	config := &FulcioConfig{
-		OIDCIssuers: map[string]OIDCIssuer{
-			"example.com": {
-				IssuerURL: "example.com",
-				ClientID:  "sigstore",
-				Type:      IssuerTypeEmail,
+	tests := []struct {
+		config *FulcioConfig
+		want   []*protobuf.OIDCIssuer
+	}{
+		{
+			config: &FulcioConfig{
+				OIDCIssuers: map[string]OIDCIssuer{
+					"example.com": {
+						IssuerURL: "example.com",
+						ClientID:  "sigstore",
+						Type:      IssuerTypeEmail,
+					},
+				},
+				MetaIssuers: map[string]OIDCIssuer{
+					"wildcard.*.example.com": {
+						ClientID: "sigstore",
+						Type:     IssuerTypeKubernetes,
+					},
+				},
+			},
+			want: []*protobuf.OIDCIssuer{
+				{
+					Audience:       "sigstore",
+					ChallengeClaim: "email",
+					Issuer: &protobuf.OIDCIssuer_IssuerUrl{
+						IssuerUrl: "example.com",
+					},
+					IssuerType: IssuerTypeEmail,
+				},
+				{
+					Audience:       "sigstore",
+					ChallengeClaim: "sub",
+					Issuer: &protobuf.OIDCIssuer_WildcardIssuerUrl{
+						WildcardIssuerUrl: "wildcard.*.example.com",
+					},
+					IssuerType: IssuerTypeKubernetes,
+				},
 			},
 		},
-		MetaIssuers: map[string]OIDCIssuer{
-			"wildcard.*.example.com": {
-				ClientID: "sigstore",
-				Type:     IssuerTypeKubernetes,
+		{
+			config: &FulcioConfig{
+				OIDCIssuers: map[string]OIDCIssuer{
+					"username.example.com": {
+						IssuerURL:     "username.example.com",
+						ClientID:      "sigstore",
+						Type:          IssuerTypeUsername,
+						SubjectDomain: "username.example.com",
+					},
+				},
+			},
+			want: []*protobuf.OIDCIssuer{
+				{
+					Audience:       "sigstore",
+					ChallengeClaim: "sub",
+					Issuer: &protobuf.OIDCIssuer_IssuerUrl{
+						IssuerUrl: "username.example.com",
+					},
+					IssuerType:    IssuerTypeUsername,
+					SubjectDomain: "username.example.com",
+				},
+			},
+		},
+		{
+			config: &FulcioConfig{
+				OIDCIssuers: map[string]OIDCIssuer{
+					"uriissuer.example.com": {
+						IssuerURL:     "uriissuer.example.com",
+						ClientID:      "sigstore",
+						Type:          IssuerTypeURI,
+						SubjectDomain: "uriissuer.example.com",
+					},
+				},
+			},
+			want: []*protobuf.OIDCIssuer{
+				{
+					Audience:       "sigstore",
+					ChallengeClaim: "sub",
+					Issuer: &protobuf.OIDCIssuer_IssuerUrl{
+						IssuerUrl: "uriissuer.example.com",
+					},
+					IssuerType:    IssuerTypeURI,
+					SubjectDomain: "uriissuer.example.com",
+				},
 			},
 		},
 	}
 
-	issuers := config.ToIssuers()
-	if len(issuers) != 2 {
-		t.Fatalf("unexpected number of issues, expected 2, got %v", len(issuers))
-	}
-
-	iss := &protobuf.OIDCIssuer{
-		Audience:       "sigstore",
-		ChallengeClaim: "email",
-		Issuer: &protobuf.OIDCIssuer_IssuerUrl{
-			IssuerUrl: "example.com",
-		},
-	}
-	if !reflect.DeepEqual(issuers[0], iss) {
-		t.Fatalf("expected issuer %v, got %v", iss, issuers[0])
-	}
-	iss = &protobuf.OIDCIssuer{
-		Audience:       "sigstore",
-		ChallengeClaim: "sub",
-		Issuer: &protobuf.OIDCIssuer_WildcardIssuerUrl{
-			WildcardIssuerUrl: "wildcard.*.example.com",
-		},
-	}
-	if !reflect.DeepEqual(issuers[1], iss) {
-		t.Fatalf("expected issuer %v, got %v", iss, issuers[1])
+	for _, test := range tests {
+		issuers := test.config.ToIssuers()
+		if !reflect.DeepEqual(issuers, test.want) {
+			t.Fatalf("expected issuers %v, got %v", test.want, issuers)
+		}
 	}
 }
 
 func TestVerifierCache(t *testing.T) {
-	cache, err := lru.New2Q(100 /* size */)
+	cache, err := lru.New2Q[string, []*verifierWithConfig](100 /* size */)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -626,6 +768,302 @@ func TestVerifierCache(t *testing.T) {
 	}
 }
 
+func TestVerifierCacheWithCustomCA(t *testing.T) {
+	caPEM, serverTLSCert, err := createServerTLS()
+	if err != nil {
+		t.Fatal(err)
+	}
+	var server *httptest.Server
+	server = httptest.NewUnstartedServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		switch r.URL.Path {
+		case "/.well-known/openid-configuration":
+			w.Header().Set("Content-Type", "application/json")
+			json.NewEncoder(w).Encode(map[string]string{
+				"issuer":   server.URL,
+				"jwks_uri": server.URL + "/keys",
+			})
+		case "/keys":
+			w.Header().Set("Content-Type", "application/json")
+			json.NewEncoder(w).Encode(map[string]interface{}{
+				"keys": []interface{}{},
+			})
+		default:
+			http.NotFound(w, r)
+		}
+	}))
+
+	server.TLS = &tls.Config{
+		Certificates: []tls.Certificate{serverTLSCert},
+	}
+	server.StartTLS()
+	defer server.Close()
+
+	cache, err := lru.New2Q[string, []*verifierWithConfig](100)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// use custom CA for OIDC issuer
+	fc := &FulcioConfig{
+		OIDCIssuers: map[string]OIDCIssuer{
+			server.URL: {
+				IssuerURL: server.URL,
+				ClientID:  "sigstore",
+				CACert:    caPEM.String(),
+			},
+		},
+		verifiers: make(map[string][]*verifierWithConfig),
+		lru:       cache,
+	}
+
+	verifier, ok := fc.GetVerifier(server.URL)
+	if !ok {
+		t.Fatal("expected to get verifier")
+	}
+	if verifier == nil {
+		t.Fatal("expected non-nil verifier")
+	}
+
+	cachedVerifier, ok := fc.GetVerifier(server.URL)
+	if !ok {
+		t.Fatal("expected to get cached verifier")
+	}
+	if !reflect.DeepEqual(verifier, cachedVerifier) {
+		t.Fatal("cached verifier doesn't match original verifier")
+	}
+
+	verifierWithOptions, ok := fc.GetVerifier(server.URL, WithSkipExpiryCheck())
+	if !ok {
+		t.Fatal("expected to get verifier with options")
+	}
+	if reflect.DeepEqual(verifier, verifierWithOptions) {
+		t.Fatal("verifier with options shouldn't match original verifier")
+	}
+}
+
+func TestVerifyK8sDefaultIssuer(t *testing.T) {
+	var (
+		crtFile, tokenFile *os.File
+		err                error
+	)
+	caPEM, serverTLSCert, err := createServerTLS()
+	if err != nil {
+		t.Fatal(err)
+	}
+	if crtFile, err = mockFile("ca-*.crt", caPEM.String()); err != nil {
+		t.Fatal(err)
+	}
+	const token = "faketoken"
+	if tokenFile, err = mockFile("token-*", token); err != nil {
+		t.Fatal(err)
+	}
+
+	var server *httptest.Server
+	counter := atomic.Uint64{}
+	authorized := atomic.Bool{}
+	server = httptest.NewUnstartedServer(http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		counter.Add(1)
+		if v := r.Header.Get("Authorization"); v == "" {
+			authorized.Store(false)
+		} else {
+			authorized.Store(true)
+			if v != fmt.Sprintf("Bearer %s", token) {
+				t.Fatal("Authorization header doesn't match the token")
+			}
+
+		}
+		switch r.URL.Path {
+		case "/.well-known/openid-configuration":
+			w.Header().Set("Content-Type", "application/json")
+			json.NewEncoder(w).Encode(map[string]string{
+				"issuer":   server.URL,
+				"jwks_uri": server.URL + "/keys",
+			})
+		case "/keys":
+			w.Header().Set("Content-Type", "application/json")
+			json.NewEncoder(w).Encode(map[string]interface{}{
+				"keys": []interface{}{},
+			})
+		default:
+			http.NotFound(w, r)
+		}
+	}))
+
+	server.TLS = &tls.Config{
+		Certificates: []tls.Certificate{serverTLSCert},
+	}
+	server.StartTLS()
+	defer server.Close()
+
+	k8sCA = crtFile.Name()
+	tokenFilePath := tokenFile.Name()
+	k8sIssuerURL = server.URL
+
+	metaIssuer := strings.ReplaceAll(server.URL, "127.0.0.1", "*.*.*.*")
+
+	tests := []struct {
+		name          string
+		fc            *FulcioConfig
+		authorized    bool
+		tokenFilePath string
+	}{
+		{
+			name:          "OIDCIssuer",
+			authorized:    true,
+			tokenFilePath: tokenFilePath,
+			fc: &FulcioConfig{
+				OIDCIssuers: map[string]OIDCIssuer{
+					k8sIssuerURL: {
+						IssuerURL: k8sIssuerURL,
+						ClientID:  "sigstore",
+						Type:      IssuerTypeKubernetes,
+					},
+				},
+			},
+		},
+		{
+			name:          "MetaIssuer",
+			authorized:    true,
+			tokenFilePath: tokenFilePath,
+			fc: &FulcioConfig{
+				MetaIssuers: map[string]OIDCIssuer{
+					metaIssuer: {
+						ClientID: "sigstore",
+						Type:     IssuerTypeKubernetes,
+					},
+				},
+			},
+		},
+		{
+			name:          "Unauthorized",
+			authorized:    false,
+			tokenFilePath: "/fake",
+			fc: &FulcioConfig{
+				OIDCIssuers: map[string]OIDCIssuer{
+					k8sIssuerURL: {
+						IssuerURL: k8sIssuerURL,
+						ClientID:  "sigstore",
+						Type:      IssuerTypeKubernetes,
+					},
+				},
+			},
+		},
+	}
+	for _, test := range tests {
+		t.Run(test.name, func(t *testing.T) {
+			k8sTokenFile = test.tokenFilePath
+			c := counter.Load()
+			if err = test.fc.prepare(); err != nil {
+				t.Fatal(err)
+			}
+
+			_, ok := test.fc.GetVerifier(k8sIssuerURL)
+			if !ok {
+				t.Fatal("expected to get verifier")
+			}
+
+			if counter.Load() <= c {
+				t.Fatal("Server API was not requested")
+			}
+
+			if authorized.Load() != test.authorized {
+				t.Fatal("Authorization header misbehaving")
+			}
+		})
+	}
+
+}
+
+func createServerTLS() (caPEM *bytes.Buffer, serverTLSCert tls.Certificate, err error) {
+	ca := &x509.Certificate{
+		SerialNumber: big.NewInt(2024),
+		Subject: pkix.Name{
+			CommonName: "Test CA",
+		},
+		NotBefore:             time.Now(),
+		NotAfter:              time.Now().AddDate(1, 0, 0),
+		IsCA:                  true,
+		KeyUsage:              x509.KeyUsageCertSign | x509.KeyUsageCRLSign,
+		BasicConstraintsValid: true,
+	}
+
+	caPrivKey, err := rsa.GenerateKey(rand.Reader, 2048)
+	if err != nil {
+		return
+	}
+
+	caBytes, err := x509.CreateCertificate(rand.Reader, ca, ca, &caPrivKey.PublicKey, caPrivKey)
+	if err != nil {
+		return
+	}
+
+	caPEM = new(bytes.Buffer)
+	pem.Encode(caPEM, &pem.Block{
+		Type:  "CERTIFICATE",
+		Bytes: caBytes,
+	})
+
+	serverCert := &x509.Certificate{
+		SerialNumber: big.NewInt(1),
+		Subject: pkix.Name{
+			CommonName: "localhost",
+		},
+		DNSNames:    []string{"localhost"},
+		IPAddresses: []net.IP{net.ParseIP("127.0.0.1")},
+		NotBefore:   time.Now(),
+		NotAfter:    time.Now().AddDate(1, 0, 0),
+		ExtKeyUsage: []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth},
+	}
+
+	serverPrivKey, err := rsa.GenerateKey(rand.Reader, 2048)
+	if err != nil {
+		return
+	}
+
+	serverCertBytes, err := x509.CreateCertificate(
+		rand.Reader,
+		serverCert,
+		ca,
+		&serverPrivKey.PublicKey,
+		caPrivKey,
+	)
+	if err != nil {
+		return
+	}
+
+	serverCertPEM := new(bytes.Buffer)
+	pem.Encode(serverCertPEM, &pem.Block{
+		Type:  "CERTIFICATE",
+		Bytes: serverCertBytes,
+	})
+
+	serverKeyPEM := new(bytes.Buffer)
+	pem.Encode(serverKeyPEM, &pem.Block{
+		Type:  "RSA PRIVATE KEY",
+		Bytes: x509.MarshalPKCS1PrivateKey(serverPrivKey),
+	})
+
+	serverTLSCert, err = tls.X509KeyPair(serverCertPEM.Bytes(), serverKeyPEM.Bytes())
+	if err != nil {
+		return
+	}
+	return
+}
+
+func mockFile(pattern, content string) (*os.File, error) {
+	tmpFile, err := os.CreateTemp("", pattern)
+	if err != nil {
+		return nil, err
+	}
+
+	_, err = tmpFile.WriteString(content)
+	if err != nil {
+		return nil, err
+	}
+	_ = tmpFile.Close()
+	return tmpFile, nil
+}
+
 type mockKeySet struct {
 }
 
