@@ -23,7 +23,6 @@ import (
 	"testing"
 
 	"github.com/tink-crypto/tink-go/v2/testutil"
-	hpkepb "github.com/tink-crypto/tink-go/v2/proto/hpke_go_proto"
 )
 
 // TODO: b/201070904 - Separate tests into internal_test package.
@@ -32,20 +31,20 @@ import (
 // https://www.rfc-editor.org/rfc/rfc9180.html#section-7.3.
 var aeadIDs = []struct {
 	name      string
-	aeadID    uint16
+	aeadID    AEADID
 	keyLength int
 }{
-	{"AES128GCM", aes128GCM, 16},
-	{"AES256GCM", aes256GCM, 32},
-	{"ChaCha20Poly1305", chaCha20Poly1305, 32},
+	{"AES128GCM", AES128GCM, 16},
+	{"AES256GCM", AES256GCM, 32},
+	{"ChaCha20Poly1305", ChaCha20Poly1305, 32},
 }
 
 type hpkeID struct {
 	id     int
 	mode   uint8
-	kemID  uint16
-	kdfID  uint16
-	aeadID uint16
+	kemID  KEMID
+	kdfID  KDFID
+	aeadID AEADID
 }
 
 type vector struct {
@@ -83,7 +82,9 @@ type encryptionString struct {
 
 type hpkeRFCTestVector struct {
 	mode                                                                                    uint8
-	kemID, kdfID, aeadID                                                                    uint16
+	kemID                                                                                   KEMID
+	kdfID                                                                                   KDFID
+	aeadID                                                                                  AEADID
 	info, pkEm, skEm, pkRm, skRm, enc, sharedSecret, keyScheduleCtx, secret, key, baseNonce string
 	consecutiveEncryptions, otherEncryptions                                                []encryptionString
 }
@@ -395,7 +396,9 @@ func aeadRFCVectors(t *testing.T) map[hpkeID]encryptionVector {
 
 	vecs := []struct {
 		mode                                              uint8
-		kemID, kdfID, aeadID                              uint16
+		kemID                                             KEMID
+		kdfID                                             KDFID
+		aeadID                                            AEADID
 		key, plaintext, associatedData, nonce, ciphertext string
 	}{
 		// https://www.rfc-editor.org/rfc/rfc9180.html#appendix-A.1.1.1
@@ -531,9 +534,9 @@ func hpkeBaseModeVectors(t *testing.T) map[hpkeID]vector {
 
 	var vecs []struct {
 		Mode             uint8             `json:"mode"`
-		KEMID            uint16            `json:"kem_id"`
-		KDFID            uint16            `json:"kdf_id"`
-		AEADID           uint16            `json:"aead_id"`
+		KEMID            KEMID             `json:"kem_id"`
+		KDFID            KDFID             `json:"kdf_id"`
+		AEADID           AEADID            `json:"aead_id"`
 		Info             testutil.HexBytes `json:"info"`
 		SenderPubKey     testutil.HexBytes `json:"pkEm"`
 		SenderPrivKey    testutil.HexBytes `json:"skEm"`
@@ -606,231 +609,3 @@ func TestLabelInfoMemoryAllocatedIsExact(t *testing.T) {
 		t.Errorf("want len(info) == cap(info), got %d != %d", len(info), cap(info))
 	}
 }
-
-func TestValidatePrivateKeyLength(t *testing.T) {
-	tests := []struct {
-		name string
-		key  *hpkepb.HpkePrivateKey
-	}{
-		{
-			name: "DHKEM_P256_HKDF_SHA256",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_P256_HKDF_SHA256,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[p256HKDFSHA256].nSK),
-			},
-		},
-		{
-			name: "DHKEM_P384_HKDF_SHA384",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_P384_HKDF_SHA384,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[p384HKDFSHA384].nSK),
-			},
-		},
-		{
-			name: "DHKEM_P521_HKDF_SHA512",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_P521_HKDF_SHA512,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[p521HKDFSHA512].nSK),
-			},
-		},
-		{
-			name: "DHKEM_X25519_HKDF_SHA256",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[x25519HKDFSHA256].nSK),
-			},
-		},
-	}
-
-	for _, test := range tests {
-		t.Run(test.name, func(t *testing.T) {
-			if err := ValidatePrivateKeyLength(test.key); err != nil {
-				t.Errorf("ValidatePrivateKeyLength(): got %v, want nil", err)
-			}
-		})
-	}
-}
-
-func TestValidatePrivateKeyLengthErrors(t *testing.T) {
-	tests := []struct {
-		name string
-		key  *hpkepb.HpkePrivateKey
-	}{
-		{
-			name: "Missing private key",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-					},
-				},
-			},
-		},
-		{
-			name: "Missing public key",
-			key: &hpkepb.HpkePrivateKey{
-				PrivateKey: []byte{},
-			},
-		},
-		{
-			name: "Zero length private key",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-					},
-				},
-				PrivateKey: []byte{},
-			},
-		},
-		{
-			name: "Wrong length private key",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[x25519HKDFSHA256].nSK+1),
-			},
-		},
-		{
-			name: "Invalid KEM",
-			key: &hpkepb.HpkePrivateKey{
-				PublicKey: &hpkepb.HpkePublicKey{
-					Params: &hpkepb.HpkeParams{
-						Kem: hpkepb.HpkeKem_KEM_UNKNOWN,
-					},
-				},
-				PrivateKey: make([]byte, kemLengths[x25519HKDFSHA256].nSK),
-			},
-		},
-	}
-
-	for _, test := range tests {
-		t.Run(test.name, func(t *testing.T) {
-			if err := ValidatePrivateKeyLength(test.key); err == nil {
-				t.Errorf("ValidatePrivateKeyLength(): got nil, want error")
-			}
-		})
-	}
-}
-
-func TestValidatePublicKeyLength(t *testing.T) {
-	tests := []struct {
-		name string
-		key  *hpkepb.HpkePublicKey
-	}{
-		{
-			name: "DHKEM_P256_HKDF_SHA256",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_P256_HKDF_SHA256,
-				},
-				PublicKey: make([]byte, kemLengths[p256HKDFSHA256].nPK),
-			},
-		},
-		{
-			name: "DHKEM_P384_HKDF_SHA384",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_P384_HKDF_SHA384,
-				},
-				PublicKey: make([]byte, kemLengths[p384HKDFSHA384].nPK),
-			},
-		},
-		{
-			name: "DHKEM_P521_HKDF_SHA512",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_P521_HKDF_SHA512,
-				},
-				PublicKey: make([]byte, kemLengths[p521HKDFSHA512].nPK),
-			},
-		},
-		{
-			name: "DHKEM_X25519_HKDF_SHA256",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-				},
-				PublicKey: make([]byte, kemLengths[x25519HKDFSHA256].nPK),
-			},
-		},
-	}
-
-	for _, test := range tests {
-		t.Run(test.name, func(t *testing.T) {
-			if err := ValidatePublicKeyLength(test.key); err != nil {
-				t.Errorf("ValidatePublicKeyLength(): got %v, want nil", err)
-			}
-		})
-	}
-}
-
-func TestValidatePublicKeyLengthErrors(t *testing.T) {
-	tests := []struct {
-		name string
-		key  *hpkepb.HpkePublicKey
-	}{
-		{
-			name: "Missing public key",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-				},
-			},
-		},
-		{
-			name: "Zero length public key",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-				},
-				PublicKey: []byte{},
-			},
-		},
-		{
-			name: "Wrong length public key",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_DHKEM_X25519_HKDF_SHA256,
-				},
-				PublicKey: make([]byte, kemLengths[x25519HKDFSHA256].nPK+1),
-			},
-		},
-		{
-			name: "Invalid KEM",
-			key: &hpkepb.HpkePublicKey{
-				Params: &hpkepb.HpkeParams{
-					Kem: hpkepb.HpkeKem_KEM_UNKNOWN,
-				},
-				PublicKey: make([]byte, kemLengths[x25519HKDFSHA256].nPK),
-			},
-		},
-	}
-
-	for _, test := range tests {
-		t.Run(test.name, func(t *testing.T) {
-			if err := ValidatePublicKeyLength(test.key); err == nil {
-				t.Errorf("ValidatePublicKeyLength(): got nil, want error")
-			}
-		})
-	}
-}
